"""Miscellaneous network and PF-related utilities"""


import re
import glob
import time
import ctypes
from socket import *
from fcntl import ioctl

from pf.constants import *
from pf.exceptions import PFError
from pf._struct import ifreq, if_data, timeval


# Dictionaries for mapping strings to constants
# Debug levels
dbg_levels = {
    "emerg":  LOG_EMERG,
    "alert":  LOG_ALERT,
    "crit":   LOG_CRIT,
    "err":    LOG_ERR,
    "warn":   LOG_WARNING,
    "notice": LOG_NOTICE,
    "info":   LOG_INFO,
    "debug":  LOG_DEBUG
}

# Memory limits
pf_limits = {
    "states":        PF_LIMIT_STATES,
    "src-nodes":     PF_LIMIT_SRC_NODES,
    "frags":         PF_LIMIT_FRAGS,
    "tables":        PF_LIMIT_TABLES,
    "table-entries": PF_LIMIT_TABLE_ENTRIES
}

# Ports, UIDs and GIDs operators
pf_ops = {
    "":   PF_OP_NONE,
    "><": PF_OP_IRG,
    "<>": PF_OP_XRG,
    "=":  PF_OP_EQ,
    "!=": PF_OP_NE,
    "<":  PF_OP_LT,
    "<=": PF_OP_LE,
    ">":  PF_OP_GT,
    ">=": PF_OP_GE,
    ":":  PF_OP_RRG
}

# Interface modifiers
pf_if_mods = {
    "network":   PFI_AFLAG_NETWORK,
    "broadcast": PFI_AFLAG_BROADCAST,
    "peer":      PFI_AFLAG_PEER,
    "0":         PFI_AFLAG_NOALIAS
}

# Global timeouts
pf_timeouts = {
    "tcp.first":       PFTM_TCP_FIRST_PACKET,
    "tcp.opening":     PFTM_TCP_OPENING,
    "tcp.established": PFTM_TCP_ESTABLISHED,
    "tcp.closing":     PFTM_TCP_CLOSING,
    "tcp.finwait":     PFTM_TCP_FIN_WAIT,
    "tcp.closed":      PFTM_TCP_CLOSED,
    "tcp.tsdiff":      PFTM_TS_DIFF,
    "udp.first":       PFTM_UDP_FIRST_PACKET,
    "udp.single":      PFTM_UDP_SINGLE,
    "udp.multiple":    PFTM_UDP_MULTIPLE,
    "icmp.first":      PFTM_ICMP_FIRST_PACKET,
    "icmp.error":      PFTM_ICMP_ERROR_REPLY,
    "other.first":     PFTM_OTHER_FIRST_PACKET,
    "other.single":    PFTM_OTHER_SINGLE,
    "other.multiple":  PFTM_OTHER_MULTIPLE,
    "frag":            PFTM_FRAG,
    "interval":        PFTM_INTERVAL,
    "adaptive.start":  PFTM_ADAPTIVE_START,
    "adaptive.end":    PFTM_ADAPTIVE_END,
    "src.track":       PFTM_SRC_NODE
}

# Syncookies modes
pf_syncookies_modes = {
    "never":    PF_SYNCOOKIES_NEVER,
    "always":   PF_SYNCOOKIES_ALWAYS,
    "adaptive": PF_SYNCOOKIES_ADAPTIVE
}

# PF Optimization Hints
pf_hint_normal = {
    "tcp.first": 2 * 60,
    "tcp.opening": 30,
    "tcp.established": 24 * 60 * 60,
    "tcp.closing": 15 * 60,
    "tcp.finwait": 45,
    "tcp.closed": 90,
    "tcp.tsdiff": 30
}

pf_hint_sattelite = {
    "tcp.first": 3 * 60,
    "tcp.opening": 30 + 5,
    "tcp.established": 24 * 60 * 60,
    "tcp.closing": 15 * 60 + 5,
    "tcp.finwait": 45 + 5,
    "tcp.closed": 90 + 5,
    "tcp.tsdiff": 60
}

pf_hint_conservative = {
    "tcp.first": 60 * 60,
    "tcp.opening": 15 * 60,
    "tcp.established": 5 * 24 * 60 * 60,
    "tcp.closing": 60 * 60,
    "tcp.finwait": 10 * 60,
    "tcp.closed": 3 * 90,
    "tcp.tsdiff": 60
}

pf_hint_aggressive = {
    "tcp.first": 30,
    "tcp.opening": 5,
    "tcp.established": 5 * 60 * 60,
    "tcp.closing": 60,
    "tcp.finwait": 30,
    "tcp.closed": 30,
    "tcp.tsdiff": 10
}

pf_hints = {
    "normal": pf_hint_normal,
    "sattelite": pf_hint_sattelite,
    "high-latency": pf_hint_sattelite,
    "conservative": pf_hint_conservative,
    "aggressive": pf_hint_aggressive
}


# Dictionaries for mapping constants to strings
# TCP states
tcpstates = {TCPS_CLOSED:       "CLOSED",
             TCPS_LISTEN:       "LISTEN",
             TCPS_SYN_SENT:     "SYN_SENT",
             TCPS_SYN_RECEIVED: "SYN_RCVD",
             TCPS_ESTABLISHED:  "ESTABLISHED",
             TCPS_CLOSE_WAIT:   "CLOSE_WAIT",
             TCPS_FIN_WAIT_1:   "FIN_WAIT_1",
             TCPS_CLOSING:      "CLOSING",
             TCPS_LAST_ACK:     "LAST_ACK",
             TCPS_FIN_WAIT_2:   "FIN_WAIT_2",
             TCPS_TIME_WAIT:    "TIME_WAIT"}

# UDP states
udpstates = {PFUDPS_NO_TRAFFIC: "NO_TRAFFIC",
             PFUDPS_SINGLE:     "SINGLE",
             PFUDPS_MULTIPLE:   "MULTIPLE"}

# ICMP and ICMPv6 codes and types
icmp_codes = {
    (ICMP_UNREACH,        ICMP_UNREACH_NET):                 "net-unr",
    (ICMP_UNREACH,        ICMP_UNREACH_HOST):                "host-unr",
    (ICMP_UNREACH,        ICMP_UNREACH_PROTOCOL):            "proto-unr",
    (ICMP_UNREACH,        ICMP_UNREACH_PORT):                "port-unr",
    (ICMP_UNREACH,        ICMP_UNREACH_NEEDFRAG):            "needfrag",
    (ICMP_UNREACH,        ICMP_UNREACH_SRCFAIL):             "srcfail",
    (ICMP_UNREACH,        ICMP_UNREACH_NET_UNKNOWN):         "net-unk",
    (ICMP_UNREACH,        ICMP_UNREACH_HOST_UNKNOWN):        "host-unk",
    (ICMP_UNREACH,        ICMP_UNREACH_ISOLATED):            "isolate",
    (ICMP_UNREACH,        ICMP_UNREACH_NET_PROHIB):          "net-prohib",
    (ICMP_UNREACH,        ICMP_UNREACH_HOST_PROHIB):         "host-prohib",
    (ICMP_UNREACH,        ICMP_UNREACH_TOSNET):              "net-tos",
    (ICMP_UNREACH,        ICMP_UNREACH_TOSHOST):             "host-tos",
    (ICMP_UNREACH,        ICMP_UNREACH_FILTER_PROHIB):       "filter-prohib",
    (ICMP_UNREACH,        ICMP_UNREACH_HOST_PRECEDENCE):     "host-preced",
    (ICMP_UNREACH,        ICMP_UNREACH_PRECEDENCE_CUTOFF):   "cutoff-preced",
    (ICMP_REDIRECT,       ICMP_REDIRECT_NET):                "redir-net",
    (ICMP_REDIRECT,       ICMP_REDIRECT_HOST):               "redir-host",
    (ICMP_REDIRECT,       ICMP_REDIRECT_TOSNET):             "redir-tos-net",
    (ICMP_REDIRECT,       ICMP_REDIRECT_TOSHOST):            "redir-tos-host",
    (ICMP_ROUTERADVERT,   ICMP_ROUTERADVERT_NORMAL):         "normal-adv",
    (ICMP_ROUTERADVERT,   ICMP_ROUTERADVERT_NOROUTE_COMMON): "common-adv",
    (ICMP_TIMXCEED,       ICMP_TIMXCEED_INTRANS):            "transit",
    (ICMP_TIMXCEED,       ICMP_TIMXCEED_REASS):              "reassemb",
    (ICMP_PARAMPROB,      ICMP_PARAMPROB_ERRATPTR):          "badhead",
    (ICMP_PARAMPROB,      ICMP_PARAMPROB_OPTABSENT):         "optmiss",
    (ICMP_PARAMPROB,      ICMP_PARAMPROB_LENGTH):            "badlen",
    (ICMP_PHOTURIS,       ICMP_PHOTURIS_UNKNOWN_INDEX):      "unknown-ind",
    (ICMP_PHOTURIS,       ICMP_PHOTURIS_AUTH_FAILED):        "auth-fail",
    (ICMP_PHOTURIS,       ICMP_PHOTURIS_DECRYPT_FAILED):     "decrypt-fail"}

icmp6_codes = {
    (ICMP6_DST_UNREACH,   ICMP6_DST_UNREACH_ADMIN):          "admin-unr",
    (ICMP6_DST_UNREACH,   ICMP6_DST_UNREACH_NOROUTE):        "noroute-unr",
    (ICMP6_DST_UNREACH,   ICMP6_DST_UNREACH_NOTNEIGHBOR):    "notnbr-unr",
    (ICMP6_DST_UNREACH,   ICMP6_DST_UNREACH_BEYONDSCOPE):    "beyond-unr",
    (ICMP6_DST_UNREACH,   ICMP6_DST_UNREACH_ADDR):           "addr-unr",
    (ICMP6_DST_UNREACH,   ICMP6_DST_UNREACH_NOPORT):         "port-unr",
    (ICMP6_TIME_EXCEEDED, ICMP6_TIME_EXCEED_TRANSIT):        "transit",
    (ICMP6_TIME_EXCEEDED, ICMP6_TIME_EXCEED_REASSEMBLY):     "reassemb",
    (ICMP6_PARAM_PROB,    ICMP6_PARAMPROB_HEADER):           "badhead",
    (ICMP6_PARAM_PROB,    ICMP6_PARAMPROB_NEXTHEADER):       "nxthdr",
    (ND_REDIRECT,         ND_REDIRECT_ONLINK):               "redironlink",
    (ND_REDIRECT,         ND_REDIRECT_ROUTER):               "redirrouter"}

icmp_types = {
    ICMP_ECHO:                  "echoreq",
    ICMP_ECHOREPLY:             "echorep",
    ICMP_UNREACH:               "unreach",
    ICMP_SOURCEQUENCH:          "squench",
    ICMP_REDIRECT:              "redir",
    ICMP_ALTHOSTADDR:           "althost",
    ICMP_ROUTERADVERT:          "routeradv",
    ICMP_ROUTERSOLICIT:         "routersol",
    ICMP_TIMXCEED:              "timex",
    ICMP_PARAMPROB:             "paramprob",
    ICMP_TSTAMP:                "timereq",
    ICMP_TSTAMPREPLY:           "timerep",
    ICMP_IREQ:                  "inforeq",
    ICMP_IREQREPLY:             "inforep",
    ICMP_MASKREQ:               "maskreq",
    ICMP_MASKREPLY:             "maskrep",
    ICMP_TRACEROUTE:            "trace",
    ICMP_DATACONVERR:           "dataconv",
    ICMP_MOBILE_REDIRECT:       "mobredir",
    ICMP_IPV6_WHEREAREYOU:      "ipv6-where",
    ICMP_IPV6_IAMHERE:          "ipv6-here",
    ICMP_MOBILE_REGREQUEST:     "mobregreq",
    ICMP_MOBILE_REGREPLY:       "mobregrep",
    ICMP_SKIP:                  "skip",
    ICMP_PHOTURIS:              "photuris"}

icmp6_types = {
    ICMP6_DST_UNREACH:          "unreach",
    ICMP6_PACKET_TOO_BIG:       "toobig",
    ICMP6_TIME_EXCEEDED:        "timex",
    ICMP6_PARAM_PROB:           "paramprob",
    ICMP6_ECHO_REQUEST:         "echoreq",
    ICMP6_ECHO_REPLY:           "echorep",
    ICMP6_MEMBERSHIP_QUERY:     "groupqry",
    MLD_LISTENER_QUERY:         "listqry",
    ICMP6_MEMBERSHIP_REPORT:    "grouprep",
    MLD_LISTENER_REPORT:        "listenrep",
    ICMP6_MEMBERSHIP_REDUCTION: "groupterm",
    MLD_LISTENER_DONE:          "listendone",
    ND_ROUTER_SOLICIT:          "routersol",
    ND_ROUTER_ADVERT:           "routeradv",
    ND_NEIGHBOR_SOLICIT:        "neighbrsol",
    ND_NEIGHBOR_ADVERT:         "neighbradv",
    ND_REDIRECT:                "redir",
    ICMP6_ROUTER_RENUMBERING:   "routrrenum",
    ICMP6_WRUREQUEST:           "wrureq",
    ICMP6_WRUREPLY:             "wrurep",
    ICMP6_FQDN_QUERY:           "fqdnreq",
    ICMP6_FQDN_REPLY:           "fqdnrep",
    ICMP6_NI_QUERY:             "niqry",
    ICMP6_NI_REPLY:             "nirep",
    MLD_MTRACE_RESP:            "mtraceresp",
    MLD_MTRACE:                 "mtrace"}


# Helper functions
def getprotobynumber(number, file="/etc/protocols"):
    """Map a protocol number to a name.

    Return the protocol name or None if no match is found.
    """
    r = re.compile("(?P<proto>\S+)\s+(?P<num>\d+)")
    with open(file, 'r') as f:
        for line in f:
            m = r.match(line)
            if m and int(m.group("num")) == number:
                return m.group("proto")


def geticmpcodebynumber(type, code, af):
    """Return the ICMP code as a string."""
    ic = icmp_codes if (af != AF_INET6) else icmp6_codes
    try:
        return ic[(type, code)]
    except KeyError:
        return None


def geticmptypebynumber(type, af):
    """Return the ICMP type as a string."""
    it = icmp_types if (af != AF_INET6) else icmp6_types
    try:
        return it[type]
    except KeyError:
        return None


def ctonm(cidr, af):
    """Convert netmask from CIDR to dotted decimal notation."""
    try:
        l = {AF_INET: 32, AF_INET6: 128}[af]
    except KeyError:
        raise ValueError("Invalid address family")

    b = "1" * cidr + "0" * (l - cidr)
    mask = "".join([chr(int(b[i:i+8], 2)) for i in range(0, l, 8)])
    mask = bytearray([int(b[i:i+8], 2) for i in range(0, l, 8)])

    return inet_ntop(af, mask)


def nmtoc(netmask, af):
    """Convert netmask from dotted decimal to CIDR notation."""
    cidr = 0
    for b in inet_pton(af, netmask):
        cidr += bin(b).count("1")
#    for b in map(ord, inet_pton(af, netmask)):
#        while b:
#            cidr += b & 1
#            b >>= 1
#
    return cidr

def is_IPaddr(addr):
    """Return True if addr is a valid IPv4 address"""
    return _is_valid_addr(AF_INET, addr)

def is_IP6addr(addr):
    """Return True if addr is a valid IPv6 address"""
    return _is_valid_addr(AF_INET6, addr)

def _is_valid_addr(af, addr):
    """Return True is addr is a valid address in the address family specified"""
    try:
        inet_pton(af, addr)
    except error:    # socket.error
        return False
    return True

def rate2str(bw):
    """Return the string representation of the network speed rate."""
    units = [" ", "K", "M", "G"]
    for unit in units:
        if bw >= 1000:
            bw /= 1000.0
        else:
            break

    if int(bw * 100 % 100):
        return "{:.2f}{}".format(bw, unit)
    else:
        return "{}{}".format(int(bw), unit)

def getifmtu(ifname):
    """Quick hack to get MTU and speed for a specified interface."""
    from pf.filter import _IOWR
    SIOCGIFMTU = _IOWR('i', 126, ifreq)
    s = socket(AF_INET, SOCK_DGRAM)
    ifrdat = if_data()
    ifr = ifreq(ifr_name=ifname, ifru_data=ctypes.addressof(ifrdat))

    try:
        ioctl(s, SIOCGIFMTU, ifr.asBuffer())
    except IOError:
        pass

    s.close()
    mtu = (ifr.ifru_metric if (ifr.ifru_metric > 0) else 1500)
    speed = ifrdat.ifi_baudrate

    return (mtu, speed)

def uptime():
    """Return system uptime in seconds"""
    CTL_KERN      = 1          # From /usr/include/sys/sysctl.h
    KERN_BOOTTIME = 21         # From /usr/include/sys/sysctl.h

    mib = (ctypes.c_int * 2)(CTL_KERN, KERN_BOOTTIME)
    tv = timeval()
    size = ctypes.c_size_t(ctypes.sizeof(timeval))

    libc = ctypes.CDLL(glob.glob("/usr/lib/libc.so*")[0], use_errno=True)
    libc.sysctl.argtypes = [ctypes.c_void_p, ctypes.c_uint, ctypes.c_void_p,
                            ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t]
    if libc.sysctl(mib, 2, ctypes.addressof(tv),
                   ctypes.addressof(size), 0, 0) == -1:
        raise PFError("Call to sysctl() failed")

    return int(time.time()) - tv.tv_sec



"""Mapping of C structs, required by ioctl() calls, to ctypes."""

from ctypes import *

from pf.constants import *


__all__ = ['timeval',
           'pfioc_limit',
           'pfioc_tm',
           'pf_status',
           'pf_addr_wrap',
           'pf_rule_addr',
           'pfsync_state_peer',
           'pfsync_state_key',
           'pfsync_state',
           'pfioc_states',
           'pfioc_state_kill',
           'pf_pool',
           'pf_rule_uid',
           'pf_rule_gid',
           'pf_threshold',
           'divert',
           'pf_rule',
           'pfioc_rule',
           'pfioc_trans_e',
           'pfioc_trans',
           'pfr_addr',
           'pfr_table',
           'pfioc_table',
           'pfr_tstats',
           'pfi_kif',
           'pfioc_iface',
           'pf_queue_bwspec',
           'pf_queue_scspec',
           'pf_queue_fqspec',
           'pf_queuespec',
           'pfioc_qstats',
           'pfioc_queue',
           'hfsc_sc',
           'hfsc_pktcntr',
           'hfsc_class_stats',
           'queue_stats',
           'ifreq',
           'if_data',
           'pfioc_synflwats']


# Constants
IFNAMSIZ             = 16               # From /usr/include/net/if.h
PFRES_MAX            = 17               # From /usr/include/net/pfvar.h
LCNT_MAX             = 10               # From /usr/include/net/pfvar.h
FCNT_MAX             = 3                # From /usr/include/net/pfvar.h
SCNT_MAX             = 3                # From /usr/include/net/pfvar.h
PF_MD5_DIGEST_LENGTH = 16               # From /usr/include/net/pfvar.h
PF_TABLE_NAME_SIZE   = 32               # From /usr/include/net/pfvar.h
PF_RULE_LABEL_SIZE   = 64               # From /usr/include/net/pfvar.h
PF_QNAME_SIZE        = 64               # From /usr/include/net/pfvar.h
PF_TAG_NAME_SIZE     = 64               # From /usr/include/net/pfvar.h
PF_SKIP_COUNT        = 9                # From /usr/include/net/pfvar.h
RTLABEL_LEN          = 32               # From /usr/include/net/route.h
PATH_MAX             = 1024             # From /usr/include/sys/syslimits.h


class pfioc_limit(Structure):     # From /usr/include/net/pfvar.h
    _fields_ = [("index",             c_int),
                ("limit",             c_uint)]


class pfioc_tm(Structure):        # From /usr/include/net/pfvar.h
    _fields_ = [("timeout",           c_int),
                ("seconds",           c_int)]


class pf_status(Structure):       # From /usr/include/net/pfvar.h
    _fields_ = [("counters",          c_uint64 * PFRES_MAX),
                ("lcounters",         c_uint64 * LCNT_MAX),
                ("fcounters",         c_uint64 * FCNT_MAX),
                ("scounters",         c_uint64 * SCNT_MAX),
                ("pcounters",         c_uint64 * 3 * 2 * 2),
                ("bcounters",         c_uint64 * 2 * 2),
                ("stateid",           c_uint64),
                ("syncookies_inflight", c_uint64 * 2),
                ("since",             c_int64),       # time_t
                ("running",           c_uint32),
                ("states",            c_uint32),
                ("states_halfopen",   c_uint32),
                ("src_nodes",         c_uint32),
                ("debug",             c_uint32),
                ("hostid",            c_uint32),
                ("reass",             c_uint32),
                ("syncookies_active", c_uint8),
                ("syncookies_mode",   c_uint8),
                ("pad",               c_uint8 * 2),
                ("ifname",            c_char * IFNAMSIZ),
                ("pf_chksum",         c_uint8 * PF_MD5_DIGEST_LENGTH)]


class pf_addr(Structure):               # From /usr/include/net/pfvar.h
    class _pfa(Union):
        _fields_ = [("v4",            c_uint32),      # struct in_addr
                    ("v6",            c_uint32 * 4),  # struct in6_addr
                    ("addr8",         c_uint8 * 16),
                    ("addr16",        c_uint16 * 8),
                    ("addr32",        c_uint32 * 4)]

    _fields_ = [("pfa",               _pfa)]
    _anonymous_ = ("pfa",)


class pf_addr_wrap(Structure):          # From /usr/include/net/pfvar.h
    class _v(Union):
        class _a(Structure):
            _fields_ = [("addr",      pf_addr),
                        ("mask",      pf_addr)]

        _fields_ = [("a",             _a),
                    ("ifname",        c_char * IFNAMSIZ),
                    ("tblname",       c_char * PF_TABLE_NAME_SIZE),
                    ("rtlabelname",   c_char * RTLABEL_LEN),
                    ("rtlabel",       c_uint32)]

    class _p(Union):
        _fields_ = [("dyn",           c_void_p),      # (struct pfi_dynaddr *)
                    ("tbl",           c_void_p),      # (struct pfr_ktable *)
                    ("dyncnt",        c_int),
                    ("tblcnt",        c_int)]

    _fields_ = [("v",                 _v),
                ("p",                 _p),
                ("type",              c_uint8),
                ("iflags",            c_uint8)]


class pf_rule_addr(Structure):          # From /usr/include/net/pfvar.h
    _fields_ = [("addr",              pf_addr_wrap),
                ("port",              c_uint16 * 2),
                ("neg",               c_uint8),
                ("port_op",           c_uint8),
                ("weight",            c_uint16)]


class pfsync_state_scrub(Structure):    # From /usr/include/net/pfvar.h
    _fields_ = [("pfss_flags",        c_uint16),
                ("pfss_ttl",          c_uint8),
                ("scrub_flag",        c_uint8),
                ("pfss_ts_mod",       c_uint32)]


class pfsync_state_peer(Structure):     # From /usr/include/net/pfvar.h
    _fields_ = [("scrub",             pfsync_state_scrub),
                ("seqlo",             c_uint32),
                ("seqhi",             c_uint32),
                ("seqdiff",           c_uint32),
                ("max_win",           c_uint16),
                ("mss",               c_uint16),
                ("state",             c_uint8),
                ("wscale",            c_uint8),
                ("pad",               c_uint8 * 6)]


class pfsync_state_key(Structure):      # From /usr/include/net/pfvar.h
    _fields_ = [("addr",              pf_addr * 2),
                ("port",              c_uint16 * 2),
                ("rdomain",           c_uint16),
                ("af",                c_uint8),       # sa_family_t
                ("pad",               c_uint8)]


class pfsync_state(Structure):          # From /usr/include/net/pfvar.h
    _fields_ = [("id",                c_uint64),
                ("ifname",            c_char * IFNAMSIZ),
                ("key",               pfsync_state_key * 2),
                ("src",               pfsync_state_peer),
                ("dst",               pfsync_state_peer),
                ("rt_addr",           pf_addr),
                ("rule",              c_uint32),
                ("anchor",            c_uint32),
                ("nat_rule",          c_uint32),
                ("creation",          c_uint32),
                ("expire",            c_uint32),
                ("packets",           c_uint32 * 2 * 2),
                ("bytes",             c_uint32 * 2 * 2),
                ("creatorid",         c_uint32),
                ("rtableid",          c_int32 * 2),
                ("max_mss",           c_uint16),
                ("af",                c_uint8),       # sa_family_t
                ("proto",             c_uint8),
                ("direction",         c_uint8),
                ("log",               c_uint8),
                ("rt",                c_uint8),
                ("timeout",           c_uint8),
                ("sync_flags",        c_uint8),
                ("updates",           c_uint8),
                ("min_ttl",           c_uint8),
                ("set_tos",           c_uint8),
                ("state_flags",       c_uint16),
                ("set_prio",          c_uint8 * 2)]


class pfioc_states(Structure):    # From /usr/include/net/pfvar.h
    class _ps_u(Union):
        _fields_ = [("ps_buf",        c_void_p),      # caddr_t
                    ("ps_states",     c_void_p)]      # struct pfsync_state *

    _fields_ = [("ps_len",            c_int),
                ("ps_u",              _ps_u)]
    _anonymous_ =  ("ps_u",)


class pf_state_cmp(Structure):          # From /usr/include/net/pfvar.h
    _fields_ = [("id",                c_uint64),
                ("creatorid",         c_uint32),
                ("direction",         c_uint8),
                ("pad",               c_uint8 * 3)]


class pfioc_state_kill(Structure): # From /usr/include/net/pfvar.h
    _fields_ = [("psk_pfcmp",         pf_state_cmp),
                ("psk_af",            c_uint8),       # sa_family_t
                ("psk_proto",         c_int),
                ("psk_src",           pf_rule_addr),
                ("psk_dst",           pf_rule_addr),
                ("psk_ifname",        c_char * IFNAMSIZ),
                ("psk_label",         c_char * PF_RULE_LABEL_SIZE),
                ("psk_killed",        c_uint),
                ("psk_rdomain",       c_uint16)]


class pf_poolhashkey(Structure):        # From /usr/include/net/pfvar.h
    class _pfk(Union):
        _fields_ = [("key8",          c_uint8 * 16),
                    ("key16",         c_uint16 * 8),
                    ("key32",         c_uint32 * 4)]

    _fields_ = [("pfk",               _pfk)]
    _anonymous_ = ("pfk",)


class pf_pool(Structure):               # From /usr/include/net/pfvar.h
    _fields_ = [("addr",              pf_addr_wrap),
                ("key",               pf_poolhashkey),
                ("counter",           pf_addr),
                ("ifname",            c_char * IFNAMSIZ),
                ("kif",               c_void_p),      # struct pfi_kif *
                ("tblidx",            c_int),
                ("states",            c_uint64),
                ("curweight",         c_int),
                ("weight",            c_uint16),
                ("proxy_port",        c_uint16 * 2),
                ("port_op",           c_uint8),
                ("opts",              c_uint8)]


class pf_rule_ptr(Union):               # From /usr/include/net/pfvar.h
    _fields_ = [("ptr",               c_void_p),      # struct pf_rule *
                ("nr",                c_uint32)]


class pf_rule_uid(Structure):           # From /usr/include/net/pfvar.h
    _fields_ = [("uid",               c_uint32 * 2),  # uid_t
                ("op",                c_uint8)]


class pf_rule_gid(Structure):           # From /usr/include/net/pfvar.h
    _fields_ = [("gid",               c_uint32 * 2),  # uid_t
                ("op",                c_uint8)]


class pf_threshold(Structure):          # From /usr/include/net/pfvar.h
    _fields_ = [("limit",             c_uint32),
                ("seconds",           c_uint32),
                ("count",             c_uint32),
                ("last",              c_uint32)]


class divert(Structure):                # From /usr/include/net/pfvar.h
    _fields_ = [("addr",              pf_addr),
                ("port",              c_uint16),
                ("type",              c_uint8)]


class pf_rule(Structure):               # From /usr/include/net/pfvar.h
    class _conn_rate(Structure):
        _fields_ = [("limit",         c_uint32),
                    ("seconds",       c_uint32)]

    _fields_ = [("src",               pf_rule_addr),
                ("dst",               pf_rule_addr),
                ("skip",              pf_rule_ptr * PF_SKIP_COUNT),
                ("label",             c_char * PF_RULE_LABEL_SIZE),
                ("ifname",            c_char * IFNAMSIZ),
                ("rcv_ifname",        c_char * IFNAMSIZ),
                ("qname",             c_char * PF_QNAME_SIZE),
                ("pqname",            c_char * PF_QNAME_SIZE),
                ("tagname",           c_char * PF_TAG_NAME_SIZE),
                ("match_tagname",     c_char * PF_TAG_NAME_SIZE),
                ("overload_tblname",  c_char * PF_TABLE_NAME_SIZE),
                ("entries",           c_void_p * 2),  # TAILQ_ENTRY(pf_rule)
                ("nat",               pf_pool),
                ("rdr",               pf_pool),
                ("route",             pf_pool),
                ("pktrate",           pf_threshold),
                ("evaluations",       c_uint64),
                ("packets",           c_uint64 * 2),
                ("bytes",             c_uint64 * 2),
                ("kif",               c_void_p),      # (struct pki_kif *)
                ("rcv_kif",           c_void_p),      # (struct pki_kif *)
                ("anchor",            c_void_p),      # (struct pf_anchor *)
                ("overload_tbl",      c_void_p),      # (struct pfr_ktable *)
                ("os_fingerprint",    c_uint32),      # pf_osfp_t
                ("rtableid",          c_int),
                ("onrdomain",         c_int),
                ("timeout",           c_uint32 * PFTM_MAX),
                ("states_cur",        c_uint32),
                ("states_tot",        c_uint32),
                ("max_states",        c_uint32),
                ("src_nodes",         c_uint32),
                ("max_src_nodes",     c_uint32),
                ("max_src_states",    c_uint32),
                ("max_src_conn",      c_uint32),
                ("max_src_conn_rate", _conn_rate),
                ("qid",               c_uint32),
                ("pqid",              c_uint32),
                ("rt_listid",         c_uint32),
                ("nr",                c_uint32),
                ("prob",              c_uint32),
                ("cuid",              c_uint32),      # uid_t
                ("cpid",              c_int32),       # pid_t
                ("return_icmp",       c_uint16),
                ("return_icmp6",      c_uint16),
                ("max_mss",           c_uint16),
                ("tag",               c_uint16),
                ("match_tag",         c_uint16),
                ("scrub_flags",       c_uint16),
                ("delay",             c_uint16),
                ("uid",               pf_rule_uid),
                ("gid",               pf_rule_gid),
                ("rule_flag",         c_uint32),
                ("action",            c_uint8),
                ("direction",         c_uint8),
                ("log",               c_uint8),
                ("logif",             c_uint8),
                ("quick",             c_uint8),
                ("ifnot",             c_uint8),
                ("match_tag_not",     c_uint8),
                ("keep_state",        c_uint8),
                ("af",                c_uint8),       # sa_family_t
                ("proto",             c_uint8),
                ("type",              c_uint16),
                ("code",              c_uint16),
                ("flags",             c_uint8),
                ("flagset",           c_uint8),
                ("min_ttl",           c_uint8),
                ("allow_opts",        c_uint8),
                ("rt",                c_uint8),
                ("return_ttl",        c_uint8),
                ("tos",               c_uint8),
                ("set_tos",           c_uint8),
                ("anchor_relative",   c_uint8),
                ("anchor_wildcard",   c_uint8),
                ("flush",             c_uint8),
                ("prio",              c_uint8),
                ("set_prio",          c_uint8 * 2),
                ("naf",               c_uint8),       # sa_family_t
                ("rcvifnot",          c_uint8),
                ("divert",            divert),
                ("exptime",           c_int64)]       # time_t


class pfioc_rule(Structure):      # From /usr/include/net/pfvar.h
    _fields_ = [("action",            c_uint32),
                ("ticket",            c_uint32),
                ("nr",                c_uint32),
                ("anchor",            c_char * PATH_MAX),
                ("anchor_call",       c_char * PATH_MAX),
                ("rule",              pf_rule)]


class pfioc_trans_e(Structure):         # From /usr/include/net/pfvar.h
    _fields_ = [("type",              c_int),
                ("anchor",            c_char * PATH_MAX),
                ("ticket",            c_uint32)]


class pfioc_trans(Structure):     # From /usr/include/net/pfvar.h
    _fields_ = [("size",              c_int),
                ("esize",             c_int),
                ("array",             c_void_p)]      # struct pfioc_trans_e *


class pfr_addr(Structure):              # From /usr/include/net/pfvar.h
    class _pfra_u(Union):
        _fields_ = [("pfra_ip4addr",  c_uint32),      # struct in_addr
                    ("pfra_ip6addr",  c_uint32 * 4)]  # struct in6_addr

    _fields_ = [("pfra_u",            _pfra_u),
                ("pfra_ifname",       c_char * IFNAMSIZ),
                ("pfra_states",       c_uint32),
                ("pfra_weight",       c_uint16),
                ("pfra_af",           c_uint8),
                ("pfra_net",          c_uint8),
                ("pfra_not",          c_uint8),
                ("pfra_fback",        c_uint8),
                ("pfra_type",         c_uint8),
                ("pad",               c_uint8 * 7)]
    _anonymous_ = ("pfra_u",)


class pfr_table(Structure):             # From /usr/include/net/pfvar.h
    _fields_ = [("pfrt_anchor",       c_char * PATH_MAX),
                ("pfrt_name",         c_char * PF_TABLE_NAME_SIZE),
                ("pfrt_flags",        c_uint32),
                ("pfrt_fback",        c_uint8)]


class pfioc_table(Structure):     # From /usr/include/net/pfvar.h
    _fields_ = [("pfrio_table",       pfr_table),
                ("pfrio_buffer",      c_void_p),
                ("pfrio_esize",       c_int),
                ("pfrio_size",        c_int),
                ("pfrio_size2",       c_int),
                ("pfrio_nadd",        c_int),
                ("pfrio_ndel",        c_int),
                ("pfrio_nchange",     c_int),
                ("pfrio_flags",       c_int),
                ("pfrio_ticket",      c_uint32)]


class pfr_tstats(Structure):            # From /usr/include/net/pfvar.h
    _fields_ = [("pfrts_t",           pfr_table),
                ("pfrts_packets",    c_uint64 * PFR_OP_TABLE_MAX * PFR_DIR_MAX),
                ("pfrts_bytes",      c_uint64 * PFR_OP_TABLE_MAX * PFR_DIR_MAX),
                ("pfrts_match",       c_uint64),
                ("pfrts_nomatch",     c_uint64),
                ("pfrts_tzero",       c_int64),         # time_t
                ("pfrts_cnt",         c_int),
                ("pfrts_refcnt",      c_int * PFR_REFCNT_MAX)]


class pfi_kif(Structure):               # From /usr/include/net/pfvar.h
    class _RB_ENTRY(Structure):
        _fields_ = [("rbe_left",      c_void_p),
                    ("rbe_right",     c_void_p),
                    ("rbe_parent",    c_void_p),
                    ("rbe_color",     c_int)]

    _fields_ = [("pfik_name",         c_char * IFNAMSIZ),
                ("pfik_tree",         _RB_ENTRY),
                ("pfik_packets",      c_uint64 * 2 * 2 * 2),
                ("pfik_bytes",        c_uint64 * 2 * 2 * 2),
                ("pfik_tzero",        c_int64),       # time_t
                ("pfik_flags",        c_int),
                ("pfik_flags_new",    c_int),
                ("pfik_ah_cookie",    c_void_p),
                ("pfik_ifp",          c_void_p),      # (struct ifnet *)
                ("pfik_group",        c_void_p),      # (struct ifg_group *)
                ("pfik_states",       c_int),
                ("pfik_rules",        c_int),
                ("pfik_routes",       c_int),
                ("pfik_srcnodes",     c_int),
                ("pfik_flagrefs",     c_int),
                ("pfik_dynaddrs",     c_void_p * 2)]  # TAILQ_HEAD(,pfi_dynaddr)


class pfioc_iface(Structure):     # From /usr/include/net/pfvar.h
    _fields_ = [("pfiio_name",        c_char * IFNAMSIZ),
                ("pfiio_buffer",      c_void_p),
                ("pfiio_esize",       c_int),
                ("pfiio_size",        c_int),
                ("pfiio_nzero",       c_int),
                ("pfiio_flags",       c_int)]


class timeval(Structure):               # From /usr/include/sys/time.h
    _fields_ = [("tv_sec",            c_int64),       # time_t
                ("tv_usec",           c_long)]        # suseconds_t


class pf_queue_bwspec(Structure):       # From /usr/include/net/pfvar.h
    _fields_ = [("absolute",          c_uint64),
                ("percent",           c_uint)]


class pf_queue_scspec(Structure):       # From /usr/include/net/pfvar.h
    _fields_ = [("m1",                pf_queue_bwspec),
                ("m2",                pf_queue_bwspec),
                ("d",                 c_uint)]


class pf_queue_fqspec(Structure):       # From /usr/include/net/pfvar.h
    _fields_ = [("flows",             c_uint),
                ("quantum",           c_uint),
                ("target",            c_uint),
                ("interval",          c_uint)]


class pf_queuespec(Structure):          # From /usr/include/net/pfvar.h
    _fields_ = [("entries",           c_void_p * 2), # TAILQ_ENTRY(pf_queuespec)
                ("qname",             c_char * PF_QNAME_SIZE),
                ("parent",            c_char * PF_QNAME_SIZE),
                ("ifname",            c_char * IFNAMSIZ),
                ("realtime",          pf_queue_scspec),
                ("linkshare",         pf_queue_scspec),
                ("upperlimit",        pf_queue_scspec),
                ("flowqueue",         pf_queue_fqspec),
                ("kif",               c_void_p),      # struct pfi_kif *
                ("flags",             c_uint),
                ("qlimit",            c_uint),
                ("qid",               c_uint32),
                ("parent_qid",        c_uint32)]


class pfioc_qstats(Structure):    # From /usr/include/net/pfvar.h
    _fields_ = [("ticket",            c_uint32),
                ("nr",                c_uint32),
                ("queue",             pf_queuespec),
                ("buf",               c_void_p),
                ("nbytes",            c_int)]


class pfioc_queue(Structure):           # From /usr/include/net/pfvar.h
    _fields_ = [("ticket",            c_uint32),
                ("nr",                c_uint),
                ("queue",             pf_queuespec)]


class hfsc_sc(Structure):               # From /usr/include/net/hfsc.h
    _fields_ = [("m1",                c_uint64),
                ("m2",                c_uint64),
                ("d",                 c_uint)]


class hfsc_pktcntr(Structure):          # From /usr/include/net/hfsc.h
    _fields_ = [("packets",           c_uint64),
                ("bytes",             c_uint64)]


class hfsc_class_stats(Structure):      # From /usr/include/net/hfsc.h
    _fields_ = [("xmit_cnt",          hfsc_pktcntr),
                ("drop_cnt",          hfsc_pktcntr),
                ("qlength",           c_uint),
                ("qlimit",            c_uint),
                ("period",            c_uint),
                ("class_id",          c_uint),
                ("class_handle",      c_uint32),
                ("rsc",               hfsc_sc),
                ("fsc",               hfsc_sc),
                ("usc",               hfsc_sc),
                ("total",             c_uint64),
                ("cumul",             c_uint64),
                ("d",                 c_uint64),
                ("e",                 c_uint64),
                ("vt",                c_uint64),
                ("f",                 c_uint64),
                ("initvt",            c_uint64),
                ("vtoff",             c_uint64),
                ("cvtmax",            c_uint64),
                ("myf",               c_uint64),
                ("cfmin",             c_uint64),
                ("cvtmin",            c_uint64),
                ("myfadj",            c_uint64),
                ("vtadj",             c_uint64),
                ("cur_time",          c_uint64),
                ("machclk_freq",      c_uint32),
                ("vtperiod",          c_uint),
                ("parentperiod",      c_uint),
                ("nactive",           c_int),
                ("qtype",             c_int)]


class queue_stats(Structure):         # From /usr/src/sbin/pfctl/pfctl_queue.c
    _fields_ = [("data",              hfsc_class_stats),
                ("avgn",              c_int),
                ("avg_bytes",         c_double),
                ("avg_packets",       c_double),
                ("prev_bytes",        c_uint64),
                ("prev_packets",      c_uint64)]


class ifreq(Structure):           # From /usr/include/net/if.h
    class _ifr_ifru(Union):
        class _sockaddr(Structure):     # From /usr/include/sys/socket.h
            _fields_ = [("sa_len",    c_uint8),
                        ("sa_family", c_uint8),       # sa_family_t
                        ("sa_data",   c_char * 14)]

        _fields_ = [("ifru_addr",     _sockaddr),
                    ("ifru_dstaddr",  _sockaddr),
                    ("ifru_broadaddr", _sockaddr),
                    ("ifru_flags",    c_short),
                    ("ifru_metric",   c_int),
                    ("ifru_vnetid",   c_int64),
                    ("ifru_media",    c_uint64),
                    ("ifru_data",     c_char_p),      # caddr_t
                    ("ifru_index",    c_uint)]

    _fields_ = [("ifr_name",          c_char * IFNAMSIZ),
                ("ifr_ifru",          _ifr_ifru)]
    _anonymous_ = ("ifr_ifru",)


class if_data(Structure):               # From /usr/include/net/if.h
    _fields_ = [("ifi_type",         c_ubyte),
                ("ifi_addrlen",      c_ubyte),
                ("ifi_hdrlen",       c_ubyte),
                ("ifi_link_state",   c_ubyte),
                ("ifi_mtu",          c_uint32),
                ("ifi_metric",       c_uint32),
                ("ifi_rdomain",      c_uint32),
                ("ifi_baudrate",     c_uint64),
                ("ifi_ipackets",     c_uint64),
                ("ifi_ierrors",      c_uint64),
                ("ifi_opackets",     c_uint64),
                ("ifi_oerrors",      c_uint64),
                ("ifi_collisions",   c_uint64),
                ("ifi_ibytes",       c_uint64),
                ("ifi_obytes",       c_uint64),
                ("ifi_imcasts",      c_uint64),
                ("ifi_omcasts",      c_uint64),
                ("ifi_iqdrops",      c_uint64),
                ("ifi_oqdrops",      c_uint64),
                ("ifi_noproto",      c_uint64),
                ("ifi_capabilities", c_uint32),
                ("ifi_lastchange",   timeval)]


class pfioc_synflwats(Structure):      # From /usr/include/net/pfvar.h
    _fields_ = [("hiwat",            c_uint32),
                ("lowat",            c_uint32)]



"""Base class for all PF-related objects"""


__all__ = ['PFObject']


class PFObject(object):
    """Base class for wrapper objects around Structures."""

    _struct_type = None

    def __init__(self, obj=None, **kwargs):
        """Check the type of obj and initialize instance attributes."""

        if self._struct_type is not None and isinstance(obj, self._struct_type):
            self._from_struct(obj)
        elif isinstance(obj, str):
            self._from_string(obj)

        self._from_kwargs(**kwargs)

    def _from_struct(self, struct):
        raise NotImplementedError()

    def _from_string(self, line):
        raise NotImplementedError()

    def _from_kwargs(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise AttributeError("Unexpected argument: {}".format(k))

    def _to_struct(self):
        raise NotImplementedError()

    def _to_string(self):
        raise NotImplementedError()

    def __str__(self):
        return self._to_string()



"""High-level classes to load rules more easily."""

import socket

from pf.rule import PFRule, PFPool, PFPort
from pf._utils import icmp_codes, icmp6_codes, icmp_types, icmp6_types
from pf.constants import *


__all__ = ["Rule",
           "BlockRule",
           "BlockInRule",
           "BlockOutRule",
           "PassRule",
           "PassInRule",
           "PassOutRule",
           "MatchRule",
           "MatchInRule",
           "MatchOutRule",
           "NATPool",
           "RDRPool",
           "TCPPort",
           "UDPPort"]


# Rules
class Rule(PFRule):
    """Generic Rule"""

    af = socket.AF_INET

    def __init__(self, **kw):
        kw.setdefault("af", self.af)

        # Try to guess protocol
        if "proto" not in kw:
            kw.setdefault("proto", 0)
            if not kw["proto"] and "src" in kw:
                kw["proto"] = kw["src"].port.proto or 0
            if not kw["proto"] and "dst" in kw:
                kw["proto"] = kw["dst"].port.proto or 0
        # Set default flags on TCP 'pass' rules
        if kw["proto"] == socket.IPPROTO_TCP and kw["action"] == PF_PASS and \
           not "flags" in kw:
            kw.update({"flags": "S", "flagset": "SA"})

        # Convert ICMP type string to constant
        if "type" in kw and isinstance(kw["type"], str):
            types = icmp6_types if (kw["af"] == socket.AF_INET6) else icmp_types
            for key, value in types.items():
                if value == kw["type"]:
                    kw["type"] = key + 1
                    break
            else:
                raise ValueError("Invalid ICMP type: {.type}".format(kw))

        # Convert ICMP code string to constants
        if "code" in kw and isinstance(kw["code"], str):
            codes = icmp6_codes if (kw["af"] == socket.AF_INET6) else icmp_codes
            for key, value in codes.items():
                if value == kw["code"]:
                    kw["type"], kw["code"] = key[0]+1, key[1]+1
                    break
            else:
                raise ValueError("Invalid ICMP code: {.code}".format(kw))

        super(Rule, self).__init__(**kw)


class BlockRule(Rule):
    """Block (drop) all traffic"""

    def __init__(self, **kw):
        super(BlockRule, self).__init__(action=PF_DROP, **kw)


class BlockInRule(BlockRule):
    """Block incoming traffic"""

    def __init__(self, **kw):
        super(BlockInRule, self).__init__(direction=PF_IN, **kw)


class BlockOutRule(BlockRule):
    """Block outgoing traffic"""

    def __init__(self, **kw):
        super(BlockOutRule, self).__init__(direction=PF_OUT, **kw)


class PassRule(Rule):
    """Pass traffic"""

    def __init__(self, **kw):
        kw.setdefault("keep_state", PF_STATE_NORMAL)
        super(PassRule, self).__init__(action=PF_PASS, **kw)


class PassInRule(PassRule):
    """Pass incoming traffic"""

    def __init__(self, **kw):
        super(PassInRule, self).__init__(direction=PF_IN, **kw)


class PassOutRule(PassRule):
    """Pass outgoing traffic"""

    def __init__(self, **kw):
        super(PassOutRule, self).__init__(direction=PF_OUT, **kw)


class MatchRule(Rule):
    """Match traffic"""

    def __init__(self, **kw):
        super(MatchRule, self).__init__(action=PF_MATCH, **kw)


class MatchInRule(MatchRule):
    """Match incoming traffic"""

    def __init__(self, **kw):
        super(MatchInRule, self).__init__(direction=PF_IN, **kw)


class MatchOutRule(MatchRule):
    """Match outgoing traffic"""

    def __init__(self, **kw):
        super(MatchOutRule, self).__init__(direction=PF_OUT, **kw)


# Pools
class NATPool(PFPool):
    """NAT address pool"""

    def __init__(self, pool, **kw):
        super(NATPool, self).__init__(PF_POOL_NAT, pool, **kw)


class RDRPool(PFPool):
    """Redirect adress pool"""

    def __init__(self, pool, **kw):
        super(RDRPool, self).__init__(PF_POOL_RDR, pool, **kw)


# Ports
class TCPPort(PFPort):
    """TCP network port"""

    def __init__(self, num, op=PF_OP_EQ):
        super(TCPPort, self).__init__(num, socket.IPPROTO_TCP, op)


class UDPPort(PFPort):
    """UDP network port"""

    def __init__(self, num, op=PF_OP_EQ):
        super(UDPPort, self).__init__(num, socket.IPPROTO_UDP, op)



"""Classes to represent Packet Filter's queueing schedulers and statistics."""

import pf._struct
from pf._base import PFObject
from pf.constants import *
from pf._utils import rate2str


__all__ = ["ServiceCurve",
           "FlowQueue",
           "PFQueue",
           "PFQueueStats"]


class ServiceCurve(PFObject):
    """ """

    _struct_type = pf._struct.pf_queue_scspec

    def __init__(self, bandwidth, burst=0, time=0):
        """ """
        if isinstance(bandwidth, pf._struct.pf_queue_scspec):
            self._from_struct(bandwidth)
        else:
            self.bandwidth = bandwidth
            self.burst = burst
            self.time = time

    def _from_struct(self, sc):
        """ """
        self.bandwidth = self._get_bandwidth(sc.m2)
        self.burst = self._get_bandwidth(sc.m1)
        self.time = sc.d

    def _to_struct(self):
        """ """
        sc = pf._struct.pf_queue_scspec()
        if (isinstance(self.bandwidth, str) and
            self.bandwidth.endswith("%")):
            sc.m2.percent = int(self.bandwidth[:-1])
        else:
            sc.m2.absolute = self.bandwidth
        if (isinstance(self.burst, str) and
            self.burst.endswith("%")):
            sc.m1.percent = int(self.burst[:-1])
        else:
            sc.m1.absolute = self.burst
        sc.d = self.time
        return sc

    def _get_bandwidth(self, bw):
        """ """
        return "{}%".format(bw.percent) if bw.percent else bw.absolute

    def _str_bandwidth(self, bw):
        """ """
        return bw if isinstance(bw, str) else rate2str(bw)

    def _to_string(self):
        """ """
        s = self._str_bandwidth(self.bandwidth)
        if self.time:
            s += " burst {}".format(self._str_bandwidth(self.burst))
            s += " for {.time}ms".format(self)

        return s


class FlowQueue(PFObject):
    """ """

    _struct_type = pf._struct.pf_queue_fqspec

    def __init__(self, flows, quantum=0, target=0, interval=0):
        """ """
        if isinstance(flows, pf._struct.pf_queue_fqspec):
            self._from_struct(flows)
        else:
            self.flows    = flows
            self.quantum  = quantum
            self.target   = target * 1000000
            self.interval = interval * 1000000

    def _from_struct(self, fq):
        """ """
        self.flows    = fq.flows
        self.quantum  = fq.quantum
        self.target   = fq.target
        self.interval = fq.interval

    def _to_struct(self):
        """ """
        fq = pf._struct.pf_queue_fqspec()
        fq.flows    = self.flows
        fq.quantum  = self.quantum
        fq.target   = self.target
        fq.interval = self.interval
        return fq

    def _to_string(self):
        """ """
        s = "flows {.flows}".format(self)
        if self.quantum:
            s += " quantum {.quantum}".format(self)
        if self.interval:
            s += " interval {}ms".format(self.interval / 1000000)
        if self.target:
            s += " target {}ms".format(self.target / 1000000)
        return s


class PFQueue(PFObject):
    """ """

    _struct_type = pf._struct.pf_queuespec

    def __init__(self, queue=None, **kw):
        """ """
        if isinstance(queue, str):
            queue = pf._struct.pf_queuespec(qname=queue, qlimit=DEFAULT_QLIMIT)
        elif queue is None:
            queue = pf._struct.pf_queuespec()
        super(PFQueue, self).__init__(queue, **kw)
        self.stats = PFQueueStats()

    def _from_struct(self, q):
        """ """
        self.qname      = q.qname.decode()
        self.parent     = q.parent.decode()
        self.ifname     = q.ifname.decode()
        self.flags      = q.flags
        self.qlimit     = q.qlimit
        self.qid        = q.qid
        self.parent_qid = q.parent_qid
        self.realtime   = ServiceCurve(q.realtime)
        self.linkshare  = ServiceCurve(q.linkshare)
        self.upperlimit = ServiceCurve(q.upperlimit)
        self.flowqueue  = FlowQueue(q.flowqueue)

    def _to_struct(self):
        """ """
        q = pf._struct.pf_queuespec()
        q.qname      = self.qname.encode()
        q.parent     = self.parent.encode()
        q.ifname     = self.ifname.encode()
        q.flags      = self.flags
        q.qlimit     = self.qlimit
        q.qid        = self.qid
        q.parent_qid = self.parent_qid
        q.realtime   = self.realtime._to_struct()
        q.linkshare  = self.linkshare._to_struct()
        q.upperlimit = self.upperlimit._to_struct()
        q.flowqueue  = self.flowqueue._to_struct()
        return q

    def _to_string(self):
        """ """
        s = "queue {.qname}".format(self)
        if self.parent and not self.parent.startswith("_"):
            s += " parent {.parent}".format(self)
        elif self.ifname:
            s += " on {.ifname}".format(self)
        if self.flags & PFQS_FLOWQUEUE:
            s += " {.flowqueue}".format(self)
        if self.linkshare.bandwidth or self.linkshare.burst:
            s += " bandwidth {}".format(self.linkshare)
        if self.realtime.bandwidth:
            s += ", min {}".format(self.realtime)
        if self.upperlimit.bandwidth:
            s += ", max {}".format(self.upperlimit)
        if self.flags & PFQS_DEFAULT:
            s += " default"
        if self.qlimit:
            s += " qlimit {.qlimit}".format(self)

        return s


class PFQueueStats(PFObject):
    """ """

    _struct_type = pf._struct.hfsc_class_stats

    def __init__(self, stats=None):
        """ """
        if stats is None:
            stats = pf._struct.hfsc_class_stats()
        super(PFQueueStats, self).__init__(stats)

    def _from_struct(self, s):
        """ """
        self.qlength = s.qlength
        self.qlimit  = s.qlimit
        self.packets = (s.xmit_cnt.packets, s.drop_cnt.packets)
        self.bytes   = (s.xmit_cnt.bytes, s.drop_cnt.bytes)

    def _to_string(self):
        """ """
        s = "  [ pkts: {0.packets[0]:10}  bytes: {0.bytes[0]:10}  "    + \
            "dropped pkts: {0.packets[1]:6} bytes: {0.bytes[1]:6} ]\n" + \
            "  [ qlength: {0.qlength:3}/{0.qlimit:3} ]"

        return s.format(self)




"""Classes to represent Packet Filter Tables."""

import re
import time
from socket import *
from ctypes import *

from pf.constants import *
from pf._struct import *
from pf._base import PFObject
from pf._utils import ctonm, nmtoc


__all__ = ['PFTableAddr',
           'PFTable',
           'PFTStats']


class PFTableAddr(PFObject):
    """Represents an address in a PF table."""

    _struct_type = pfr_addr

    def __init__(self, addr=None, **kw):
        """Check argument and initialize class attributes."""
        if addr is None:
            addr = pfr_addr()

        super(PFTableAddr, self).__init__(addr, **kw)

    def _from_struct(self, a):
        """Initialize class attributes from a pfr_addr structure."""
        l = {AF_INET: 4, AF_INET6: 16}[a.pfra_af]

        self.af     = a.pfra_af
        self.addr   = inet_ntop(self.af, string_at(addressof(a.pfra_u), l))
        self.mask   = ctonm(a.pfra_net, self.af)
        self.neg    = bool(a.pfra_not)
        self.fback  = a.pfra_fback
        self.ifname = a.pfra_ifname.decode()
        self.type   = a.pfra_type
        self.states = a.pfra_states
        self.weight = a.pfra_weight

    def _from_string(self, a):
        """Initalize a new instance from a string."""
        ipv4_re = "[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"
        ipv6_re = "[0-9a-f:]+"
        addr_re = "(?P<neg>!)?\s*"                            + \
                  "(?P<address>(?P<ipv4>{})|".format(ipv4_re) + \
                              "(?P<ipv6>{}))".format(ipv6_re) + \
                              "(?:/(?P<mask>\d+))?\s*"

        m = re.compile(addr_re).match(a)
        if not m:
            raise ValueError("Could not parse address: '{}'".format(a))

        self.neg = bool(m.group("neg"))

        if m.group("ipv4"):
            self.af = AF_INET
            self.addr = m.group("ipv4")
        elif m.group("ipv6"):
            self.af = AF_INET6
            self.addr = m.group("ipv6")

        net = m.group("mask") or {AF_INET: 32, AF_INET6: 128}[self.af]
        self.mask = ctonm(int(net), self.af)

        self.fback = 0
        self.ifname = ""               # ?
        self.type   = PFRKE_PLAIN      # ?
        self.states = 0
        self.weight = 0

    def _to_struct(self):
        """Convert this instance to a pfr_addr structure."""
        a = pfr_addr()

        addr = inet_pton(self.af, self.addr)
        memmove(a.pfra_ip6addr, c_char_p(addr), len(addr))

        a.pfra_af     = self.af
        a.pfra_net    = nmtoc(self.mask, self.af)
        a.pfra_not    = int(self.neg)
        a.pfra_fback  = self.fback
        a.pfra_ifname = self.ifname.encode()
        a.pfra_type   = self.type

        return a

    def _to_string(self):
        """Return the string representation of the address."""
        s = ("! {!s}" if self.neg else "{!s}").format(self.addr)
        bits = nmtoc(self.mask, self.af)
        if not ((self.af == AF_INET and bits == 32) or (bits == 128)):
            s += "/{}".format(bits)

        return s


class PFTable(PFObject):
    """Represents a PF table."""

    _struct_type = pfr_table

    def __init__(self, table=None, *addrs, **kw):
        """Check argument and initialize class attributes."""
        if table is None:
            table = pfr_table()
        elif isinstance(table, str):
            table = pfr_table(pfrt_name=table.encode())

        self._addrs = []
        for addr in addrs:
            if not isinstance(addr, PFTableAddr):
                addr = PFTableAddr(addr)
            self._addrs.append(addr)

        super(PFTable, self).__init__(table, **kw)

    @property
    def addrs(self):
        """Return a tuple containing the address in the table."""
        return tuple(self._addrs)

    def _from_struct(self, t):
        """Initialize class attributes from a pfr_table structure"""
        self.anchor = t.pfrt_anchor.decode()
        self.name   = t.pfrt_name.decode()
        self.flags  = t.pfrt_flags
        self.fback  = t.pfrt_fback

    def _to_struct(self):
        """Convert this instance to a pfr_table structure."""
        t = pfr_table()

        t.pfrt_anchor = self.anchor.encode()
        t.pfrt_name   = self.name.encode()
        t.pfrt_flags  = self.flags & (PFR_TFLAG_CONST|PFR_TFLAG_PERSIST)
        t.pfrt_fback  = self.fback

        return t

    def _to_string(self):
        """Return the string representation of the table."""
        s  = ('c' if (self.flags & PFR_TFLAG_CONST) else '-')
        s += ('p' if (self.flags & PFR_TFLAG_PERSIST) else '-')
        s += ('a' if (self.flags & PFR_TFLAG_ACTIVE) else '-')
        s += ('i' if (self.flags & PFR_TFLAG_INACTIVE) else '-')
        s += ('r' if (self.flags & PFR_TFLAG_REFERENCED) else '-')
        s += ('h' if (self.flags & PFR_TFLAG_REFDANCHOR) else '-')
        s += ('C' if (self.flags & PFR_TFLAG_COUNTERS) else '-')
        s += "\t{.name}".format(self)

        if self.anchor:
            s += "\t{.anchor}".format(self)

        return s


class PFTStats(PFObject):
    """Class containing statistics for a PF table."""

    _struct_type = pfr_tstats

    def __init__(self, tstats):
        """Initialize class attributes."""
        super(PFTStats, self).__init__(tstats)

    def _from_struct(self, s):
        """Initialize class attributes from a pfr_tstats structure."""
        self.table   = PFTable(s.pfrts_t)
        self.packets = {"in":  tuple(s.pfrts_packets[PFR_DIR_IN]),
                        "out": tuple(s.pfrts_packets[PFR_DIR_OUT])}
        self.bytes   = {"in":  tuple(s.pfrts_bytes[PFR_DIR_IN]),
                        "out": tuple(s.pfrts_bytes[PFR_DIR_OUT])}
        self.cleared = s.pfrts_tzero
        self.cnt     = s.pfrts_cnt
        self.evalcnt = {"match":   s.pfrts_match,
                        "nomatch": s.pfrts_nomatch}
        self.refcnt  = {"rules":   s.pfrts_refcnt[PFR_REFCNT_RULE],
                        "anchors": s.pfrts_refcnt[PFR_REFCNT_ANCHOR]}

    def _to_string(self):
        """Return the string representation of the table statistics."""
        s  = "{.table}\n".format(self)
        s += "\tAddresses:   {.cnt:d}\n".format(self)
        s += "\tCleared:     {}\n".format(time.ctime(self.cleared))
        s += "\tReferences:  [ Anchors: {anchors:<18d} Rules: {rules:<18d} ]\n"
        s += "\tEvaluations: [ NoMatch: {nomatch:<18d} Match: {match:<18d} ]\n"
        s = s.format(**dict(self.refcnt, **self.evalcnt))

        pfr_ops = ("Block:", "Pass:", "XPass:")
        for o, p, b in zip(pfr_ops, self.packets["in"], self.bytes["in"]):
            l = "\tIn/{:<6s}    [ Packets: {:<18d} Bytes: {:<18d} ]\n"
            s += l.format(o, p, b)
        for o, p, b in zip(pfr_ops, self.packets["out"], self.bytes["out"]):
            l = "\tOut/{:<6s}   [ Packets: {:<18d} Bytes: {:<18d} ]\n"
            s += l.format(o, p, b)

        return s.rstrip()



"""A class for managing OpenBSD's Packet Filter.

This class communicates with the kernel through the ioctl(2) interface provided
by the pf(4) pseudo-device; this allows Python to natively send commands to the
kernel, thanks to the fcntl and ctypes modules.
"""

import os
import stat
from fcntl import ioctl
from errno import *
from ctypes import *
from socket import *

from pf.exceptions import PFError
from pf.constants import *
from pf._struct import *
from pf._base import PFObject
from pf.queue import *
from pf.state import PFState
from pf.status import PFStatus, PFIface
from pf.table import PFTableAddr, PFTable, PFTStats
from pf.rule import PFRule, PFRuleset, pf_timeouts
from pf._utils import *


__all__ = ['PacketFilter']


# ioctl() operations
IOCPARM_MASK     = 0x1fff
IOC_VOID         = 0x20000000
IOC_OUT          = 0x40000000
IOC_IN           = 0x80000000
IOC_INOUT        = IOC_IN | IOC_OUT

def _IOC(inout, group, num, len):
    return (inout | ((len & IOCPARM_MASK) << 16) | ((group) << 8) | (num))

def _IO(group, num):
    return _IOC(IOC_VOID, ord(group), num, 0)

def _IOWR(group, num, type):
    return _IOC(IOC_INOUT, ord(group), num, sizeof(type))

DIOCSTART        = _IO  ('D',  1)
DIOCSTOP         = _IO  ('D',  2)
DIOCADDRULE      = _IOWR('D',  4, pfioc_rule)
DIOCGETRULES     = _IOWR('D',  6, pfioc_rule)
DIOCGETRULE      = _IOWR('D',  7, pfioc_rule)
DIOCCLRSTATES    = _IOWR('D', 18, pfioc_state_kill)
#DIOCGETSTATE     = _IOWR('D', 19, pfioc_state)
DIOCSETSTATUSIF  = _IOWR('D', 20, pfioc_iface)
DIOCGETSTATUS    = _IOWR('D', 21, pf_status)
DIOCCLRSTATUS    = _IOWR('D', 22, pfioc_iface)
#DIOCNATLOOK      = _IOWR('D', 23, pfioc_natlook)
DIOCSETDEBUG     = _IOWR('D', 24, c_uint32)
DIOCGETSTATES    = _IOWR('D', 25, pfioc_states)
#DIOCCHANGERULE   = _IOWR('D', 26, pfioc_rule)
DIOCSETTIMEOUT   = _IOWR('D', 29, pfioc_tm)
DIOCGETTIMEOUT   = _IOWR('D', 30, pfioc_tm)
#DIOCADDSTATE     = _IOWR('D', 37, pfioc_state)
#DIOCCLRRULECTRS  = _IO  ('D', 38)
DIOCGETLIMIT     = _IOWR('D', 39, pfioc_limit)
DIOCSETLIMIT     = _IOWR('D', 40, pfioc_limit)
DIOCKILLSTATES   = _IOWR('D', 41, pfioc_state_kill)
#DIOCGETRULESETS  = _IOWR('D', 58, pfioc_ruleset)
#DIOCGETRULESET   = _IOWR('D', 59, pfioc_ruleset)
DIOCRCLRTABLES   = _IOWR('D', 60, pfioc_table)
DIOCRADDTABLES   = _IOWR('D', 61, pfioc_table)
DIOCRDELTABLES   = _IOWR('D', 62, pfioc_table)
DIOCRGETTABLES   = _IOWR('D', 63, pfioc_table)
DIOCRGETTSTATS   = _IOWR('D', 64, pfioc_table)
DIOCRCLRTSTATS   = _IOWR('D', 65, pfioc_table)
DIOCRCLRADDRS    = _IOWR('D', 66, pfioc_table)
DIOCRADDADDRS    = _IOWR('D', 67, pfioc_table)
DIOCRDELADDRS    = _IOWR('D', 68, pfioc_table)
DIOCRSETADDRS    = _IOWR('D', 69, pfioc_table)
DIOCRGETADDRS    = _IOWR('D', 70, pfioc_table)
#DIOCRGETASTATS   = _IOWR('D', 71, pfioc_table)
#DIOCRCLRASTATS   = _IOWR('D', 72, pfioc_table)
DIOCRTSTADDRS    = _IOWR('D', 73, pfioc_table)
#DIOCRSETTFLAGS   = _IOWR('D', 74, pfioc_table)
DIOCRINADEFINE   = _IOWR('D', 77, pfioc_table)
#DIOCOSFPFLUSH    = _IO  ('D', 78)
#DIOCOSFPADD      = _IOWR('D', 79, pf_osfp_ioctl)
#DIOCOSFPGET      = _IOWR('D', 80, pf_osfp_ioctl)
DIOCXBEGIN       = _IOWR('D', 81, pfioc_trans)
DIOCXCOMMIT      = _IOWR('D', 82, pfioc_trans)
DIOCXROLLBACK    = _IOWR('D', 83, pfioc_trans)
#DIOCGETSRCNODES  = _IOWR('D', 84, pfioc_src_nodes)
#DIOCCLRSRCNODES  = _IO  ('D', 85)
DIOCSETHOSTID    = _IOWR('D', 86, c_uint32)
DIOCIGETIFACES   = _IOWR('D', 87, pfioc_iface)
DIOCSETIFFLAG    = _IOWR('D', 89, pfioc_iface)
DIOCCLRIFFLAG    = _IOWR('D', 90, pfioc_iface)
#DIOCKILLSRCNODES = _IOWR('D', 91, pfioc_src_node_kill)
DIOCSETREASS     = _IOWR('D', 92, c_uint32)
DIOCADDQUEUE	 = _IOWR('D', 93, pfioc_queue)
DIOCGETQUEUES	 = _IOWR('D', 94, pfioc_queue)
DIOCGETQUEUE	 = _IOWR('D', 95, pfioc_queue)
DIOCGETQSTATS	 = _IOWR('D', 96, pfioc_qstats)
DIOCSETSYNFLWATS  = _IOWR('D', 97, pfioc_synflwats)
DIOCSETSYNCOOKIES = _IOWR('D', 98, c_uint8)
DIOCGETSYNFLWATS  = _IOWR('D', 99, pfioc_synflwats)



class _PFTrans(object):
    """Class for managing transactions with the Packet Filter subsystem."""

    def __init__(self, dev, path="", *trans_type):
        """Initialize the required structures."""
        self.dev = dev
        self.size = len(trans_type)
        self.array = (pfioc_trans_e * self.size)()

        for a, t in zip(self.array, trans_type):
            a.type = t
            a.anchor = path.encode()

        self._pt = pfioc_trans(size=self.size, esize=sizeof(pfioc_trans_e),
                               array=addressof(self.array))

    def __enter__(self):
        """Start the transaction."""
        ioctl(self.dev, DIOCXBEGIN, self._pt)
        return self

    def __exit__(self, type, value, traceback):
        """Commit changes if no exceptions occurred; otherwise, rollback."""
        if type is None:
            ioctl(self.dev, DIOCXCOMMIT, self._pt)
        else:
            ioctl(self.dev, DIOCXROLLBACK, self._pt)


class PacketFilter(object):
    """Class representing the kernel's packet filtering subsystem.

    It provides a set of methods that allow you to send commands to the kernel
    through the ioctl(2) interface provided by the pf(4) pseudo-device.
    Basically, all methods in this class are just wrappers to ioctl(2) calls,
    and may consequently raise IOError if the ioctl() request fails.
    """

    def __init__(self, dev="/dev/pf"):
        """Set the pf device."""
        self.dev = dev

    def enable(self):
        """Enable Packet Filtering."""
        with open(self.dev, 'w') as d:
            try:
                ioctl(d, DIOCSTART)
            except IOError as e:
                if e != EEXIST:       # EEXIST means PF is already enabled
                    raise

    def disable(self):
        """Disable Packet Filtering."""
        with open(self.dev, 'w') as d:
            try:
                ioctl(d, DIOCSTOP)
            except IOError as e:
                if e != ENOENT:       # ENOENT means PF is already disabled
                    raise 

    def set_debug(self, level):
        """Set the debug level.

        The debug level can be either one of the LOG_* constants or a string.
        """
        if level in dbg_levels:
            level = dbg_levels[level]

        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                ioctl(d, DIOCSETDEBUG, c_uint32(level))

    def set_hostid(self, id):
        """Set the host ID.

        The host ID is used by pfsync to identify the host that created a state
        table entry. 'id' must be a 32-bit unsigned integer.
        """
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                ioctl(d, DIOCSETHOSTID, c_uint32(htonl(id)))

    def set_reassembly(self, reassembly):
        """Enable reassembly of network traffic.

        The 'reassembly' argument specifies the flags for the reassembly
        operation; available flags are PF_REASS_ENABLED and PF_REASS_NODF.
        """
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                ioctl(d, DIOCSETREASS, c_uint32(reassembly))

    def get_limit(self, limit=None):
        """Return the hard limits on the memory pools used by Packet Filter.

        'limit' can be either one of the PF_LIMIT_* constants or a string;
        return the value of the requested limit (UINT_MAX means unlimited) or,
        if called with no arguments, a dictionary containing all the available
        limits.
        """
        if limit is None:
            return dict([(l, self.get_limit(l)) for l in pf_limits])
        elif limit in pf_limits:
            limit = pf_limits[limit]

        pl = pfioc_limit(index=limit)

        with open(self.dev, 'r') as d:
            ioctl(d, DIOCGETLIMIT, pl)

        return pl.limit

    def set_limit(self, limit, value):
        """Set hard limits on the memory pools used by Packet Filter.

        'limit' can be either one of the PF_LIMIT_* constants or a string; a
        'value' of UINT_MAX means unlimited. Raise PFError if the current pool
        size exceeds the requested hard limit.
        """
        if limit in pf_limits:
            limit = pf_limits[limit]

        pl = pfioc_limit(index=limit, limit=value)
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                try:
                    ioctl(d, DIOCSETLIMIT, pl)
                except IOError as e:
                    if e == EBUSY:
                        raise PFError("Current pool size > {0:d}".format(value))
                    raise

    def get_timeout(self, timeout=None):
        """Return the configured timeout values for PF states.

        'timeout' can be either one of the PFTM_* constants or a string; return
        the value of the requested timeout or, if called with no arguments, a
        dictionary containing all the available timeouts.
        """
        if timeout is None:
            return dict([(t, self.get_timeout(t)) for t in pf_timeouts])
        elif timeout in pf_timeouts:
            timeout = pf_timeouts[timeout]

        tm = pfioc_tm(timeout=timeout)
        with open(self.dev, 'r') as d:
            ioctl(d, DIOCGETTIMEOUT, tm)

        return tm.seconds

    def set_timeout(self, timeout, value):
        """Set the timeout 'value' for a specific PF state.

        'timeout' can be either one of the PFTM_* constants or a string; return
        the old value of the specified timeout.
        """
        if timeout in pf_timeouts:
            timeout = pf_timeouts[timeout]

        tm = pfioc_tm(timeout=timeout, seconds=value)
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                ioctl(d, DIOCSETTIMEOUT, tm)

        return tm.seconds

    def set_optimization(self, opt="normal"):
        """Set the optimization profile for state handling like pfctl."""
        for name, val in pf_hints[opt].items():
            self.set_timeout(name, val)

    def get_optimization(self):
        """ """
        tm = self.get_timeout()
        for name, val in pf_hints.items():
            if val["tcp.first"] == tm["tcp.first"]:
                return name

    def get_ifaces(self, ifname="", size=8):
        """Get the list of interfaces and interface drivers known to pf.

        Return a tuple of PFIface objects or a single PFIface object if a
        specific 'ifname' is specified.
        """
        pi = pfioc_iface(pfiio_name=ifname.encode(),
                         pfiio_esize=sizeof(pfi_kif),
                         pfiio_size=size+1)

        with open(self.dev, 'w') as d:
            buf = (pfi_kif * pi.pfiio_size)()
            pi.pfiio_buffer = addressof(buf)
            ioctl(d, DIOCIGETIFACES, pi)

        if ifname:
            return PFIface(buf[0])
        else:
            ifaces = tuple(i for i in map(PFIface, buf) if i.name)
            if len(ifaces) == size:
                return self.get_ifaces(ifname, size*2)
            return ifaces

    def set_ifflags(self, ifname, flags):
        """Set the user setable 'flags' on the interface 'ifname'."""
        pi = pfioc_iface(pfiio_name=ifname.encode(), pfiio_flags=flags)
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                ioctl(d, DIOCSETIFFLAG, pi)

    def clear_ifflags(self, ifname, flags=None):
        """Clear the specified user setable 'flags' on the interface 'ifname'.

        If no flags are specified, clear all flags.
        """
        if flags is None:
            flags = PFI_IFLAG_SKIP

        pi = pfioc_iface(pfiio_name=ifname.encode(), pfiio_flags=flags)
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                ioctl(d, DIOCCLRIFFLAG, pi)

    def set_status_if(self, ifname=""):
        """Specify the interface for which statistics are accumulated.

        If no 'ifname' is provided, turn off the collection of per-interface
        statistics. Raise PFError if 'ifname' is not a valid interface name.
        """
        pi = pfioc_iface(pfiio_name=ifname.encode())
        with open(self.dev, 'w') as d:
            with _PFTrans(d):
                try:
                    ioctl(d, DIOCSETSTATUSIF, pi)
                except IOError as e:
                    if e == EINVAL:
                        raise PFError("Invalid ifname: '{0}'".format(ifname))
                    raise

    def get_status(self):
        """Return a PFStatus object containing the internal PF statistics."""
        s = pf_status()
        with open(self.dev, 'w') as d:
            ioctl(d, DIOCGETSTATUS, s)

        return PFStatus(s)

    def clear_status(self, ifname=""):
        """Clear the internal packet filter statistics.

        An optional 'ifname' can be specified in order to clear statistics only
        for a specific interface.
        """
        pi = pfioc_iface(pfiio_name=ifname.encode())
        with open(self.dev, 'w') as d:
            ioctl(d, DIOCCLRSTATUS, pi)

    def get_states(self):
        """Retrieve Packet Filter's state table entries.

        Return a tuple of PFState objects representing the states currently
        tracked by PF.
        """
        ps = pfioc_states()

        l  = 0
        with open(self.dev, 'w') as d:
            while True:
                if l:
                    ps_states = (pfsync_state * int(l / sizeof(pfsync_state)))()
                    ps.ps_buf = addressof(ps_states)
                    ps.ps_len = l
                ioctl(d, DIOCGETSTATES, ps)
                if ps.ps_len == 0:
                    return ()
                if ps.ps_len <= l:
                    break
                l = (ps.ps_len * 2)

        ps_num = int(ps.ps_len / sizeof(pfsync_state))
        return tuple([PFState(s) for s in ps_states[:ps_num]])

    def clear_states(self, ifname=""):
        """Clear all states.

        If an interface name is provided, only states for that interface will
        be cleared. Return the number of cleared states.
        """
        psk = pfioc_state_kill(psk_ifname=ifname)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCCLRSTATES, psk)

        return psk.psk_killed

    def kill_states(self, af=AF_UNSPEC, proto=0, src=None, dst=None, ifname="",
                    label="", rdomain=0):
        """Clear states matching the specified arguments.

        States can be specified by address family, layer-4 protocol, source and
        destination addresses, interface name, label and routing domain. Return
        the number of killed states.
        """
        psk = pfioc_state_kill(psk_af=af, psk_proto=proto, psk_ifname=ifname,
                               psk_label=label, psk_rdomain=rdomain)
        if src:
            psk.psk_src = src._to_struct()
        if dst:
            psk.psk_dst = dst._to_struct()

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCKILLSTATES, psk)

        return psk.psk_killed

    def clear_rules(self, path=""):
        """Clear all rules contained in the anchor 'path'."""
        self.load_ruleset(PFRuleset(), path, PF_TRANS_RULESET)

    def load_queues(self, *queues):
        """Load a set of queues on an interface.

        'queues' must be PFQueue objects.
        """
        with open(self.dev, 'w') as d:
            with _PFTrans(d, "", PF_TRANS_RULESET) as t:
                for queue in queues:
                    q = pfioc_queue(ticket=t.array[0].ticket,
                                    queue=queue._to_struct())
                    ioctl(d, DIOCADDQUEUE, q)

    def get_queues(self):
        """Retrieve the currently loaded queues.

        Return a tuple of PFQueue objects.
        """
        queues = []
        pq = pfioc_queue()
        with open(self.dev, 'r') as d:
            ioctl(d, DIOCGETQUEUES, pq)

            qstats = queue_stats()
            for nr in range(pq.nr):
                pqs = pfioc_qstats(nr=nr, ticket=pq.ticket,
                                   buf=addressof(qstats.data),
                                   nbytes=sizeof(hfsc_class_stats))
                ioctl(d, DIOCGETQSTATS, pqs)
                queue = PFQueue(pqs.queue)
                queue.stats = PFQueueStats(qstats.data)
                queues.append(queue)

        return queues

    def _get_rules(self, path, dev, clear):
        """Recursively retrieve rules from the specified ruleset."""
        if path.endswith("/*"):
            path = path[:-2]

        pr = pfioc_rule(anchor=path.encode())
        if clear:
            pr.action = PF_GET_CLR_CNTR

        pr.rule.action = PF_PASS
        ioctl(dev, DIOCGETRULES, pr)

        tables = list(self.get_tables(PFTable(anchor=path)))
        rules = []
        for nr in range(pr.nr):
            pr.nr = nr
            ioctl(dev, DIOCGETRULE, pr)
            if pr.anchor_call:
                path = os.path.join(pr.anchor.decode(), pr.anchor_call.decode())
                rs = PFRuleset(pr.anchor_call.decode(), pr.rule)
                rs.append(*self._get_rules(path, dev, clear))
                rules.append(rs)
            else:
                rules.append(PFRule(pr.rule))

        return tables + rules

    def get_ruleset(self, path="", clear=False, **kw):
        """Return a PFRuleset object containing the active ruleset.
        
        'path' is the path of the anchor to retrieve rules from. If 'clear' is
        True, per-rule statistics will be cleared. Keyword arguments can be
        passed for returning only matching rules.
        """
        rs = PFRuleset(os.path.basename(path))

        with open(self.dev, 'r') as d:
            for rule in self._get_rules(path, d, clear):
                if isinstance(rule, PFRule):
                    if not all((getattr(rule, attr) == value)
                               for (attr, value) in kw.items()):
                        continue
                rs.append(rule)
        return rs

    def _inadefine(self, table, dev, path, ticket):
        """Define a table in the inactive ruleset."""
        table.anchor = path
        io = pfioc_table(pfrio_table=table._to_struct(), pfrio_ticket=ticket,
                         pfrio_esize=sizeof(pfr_addr))

        if table.addrs:
            io.pfrio_flags |= PFR_FLAG_ADDRSTOO
            addrs = table.addrs
            buf = (pfr_addr * len(addrs))(*[a._to_struct() for a in addrs])
            io.pfrio_buffer = addressof(buf)
            io.pfrio_size = len(addrs)

        ioctl(dev, DIOCRINADEFINE, io)

    def load_ruleset(self, ruleset, path="", *tr_type):
        """Load the given ruleset.

        'ruleset' must be a PFRuleset object; 'path' is the name of the anchor
        where to load rules; 'tr_type' is one or more PF_TRANS_* constants: if
        omitted, all ruleset types will be loaded.
        """
        if not tr_type:
            tr_type = (PF_TRANS_TABLE, PF_TRANS_RULESET)

        with open(self.dev, 'w') as d:
            with _PFTrans(d, path, *tr_type) as t:
                for a in t.array:
                    if a.type == PF_TRANS_TABLE:
                        for t in ruleset.tables:
                            self._inadefine(t, d, path, a.ticket)
                    elif a.type == PF_TRANS_RULESET:
                        for r in ruleset.rules:
                            pr = pfioc_rule(ticket=a.ticket,
                                            anchor=path.encode(),
                                            rule=r._to_struct())

                            if isinstance(r, PFRuleset):
                                pr.anchor_call = r.name.encode()

                            ioctl(d, DIOCADDRULE, pr)

                            if isinstance(r, PFRuleset):
                                self.load_ruleset(r, os.path.join(path, r.name),
                                                  *tr_type)

    def add_tables(self, *tables):
        """Create one or more tables.

        'tables' must be PFTable objects; return the number of tables created.
        """
        io = pfioc_table(pfrio_esize=sizeof(pfr_table), pfrio_size=len(tables))

        buffer = (pfr_table * len(tables))(*[t._to_struct() for t in tables])
        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRADDTABLES, io)

        for t in filter(lambda t: t.addrs, tables):
            self.add_addrs(t, *t.addrs)

        return io.pfrio_nadd

    def clear_tables(self, filter=None):
        """Clear all tables.

        'filter' is a PFTable object that allows you to specify the anchor of
        the tables to delete. Return the number of tables deleted.
        """
        io = pfioc_table()

        if filter is not None:
            io.pfrio_table = pfr_table(pfrt_anchor=filter.anchor)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRCLRTABLES, io)

        return io.pfrio_ndel

    def del_tables(self, *tables):
        """Delete one or more tables.

        'tables' must be PFTable objects. Return the number of tables deleted.
        """
        io = pfioc_table(pfrio_esize=sizeof(pfr_table), pfrio_size=len(tables))

        buffer = (pfr_table * len(tables))()
        for (t, b) in zip(tables, buffer):
            b.pfrt_name = t.name.encode()
            b.pfrt_anchor = t.anchor.encode()

        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRDELTABLES, io)

        return io.pfrio_ndel

    def get_tables(self, filter=None, buf_size=10):
        """Get the list of all tables.

        'filter' is a PFTable object that allows you to specify the anchor of
        the tables to retrieve. Return a tuple of PFTable objects containing
        the currently-loaded tables.
        """
        io = pfioc_table(pfrio_esize=sizeof(pfr_table))

        if filter is not None:
            io.pfrio_table = pfr_table(pfrt_anchor=filter.anchor.encode())

        with open(self.dev, 'w') as d:
            while True:
                buffer = (pfr_table * buf_size)()
                io.pfrio_buffer = addressof(buffer)
                io.pfrio_size = buf_size

                ioctl(d, DIOCRGETTABLES, io)

                if io.pfrio_size <= buf_size:
                    break
                buf_size = io.pfrio_size

        tables = []
        for t in buffer[:io.pfrio_size]:
            try:
                addrs = self.get_addrs(PFTable(t))
            except IOError as e:
                pass       # Ignore tables of which you can't get the addresses
            else:
                tables.append(PFTable(t, *addrs))

        return tuple(tables)

    def test_addrs(self, table, *addrs):
        """Test if one or more addresses match a table.

        'table' can be either a PFTable instance or a string containing the
        table name; 'addrs' can be either PFTableAddr instances or strings.
        Return the addresses that match.
        """
        if isinstance(table, str):
            table = pfr_table(pfrt_name=table.encode())
        else:
            table = pfr_table(pfrt_name=table.name.encode(),
                              pfrt_anchor=table.anchor.encode())

        _addrs = []
        for addr in addrs:
            if isinstance(addr, PFTableAddr):
                _addrs.append(addr)
            else:
                _addrs.append(PFTableAddr(addr))

        io = pfioc_table(pfrio_table=table, pfrio_esize=sizeof(pfr_addr),
                         pfrio_size=len(addrs))

        buffer = (pfr_addr * len(addrs))(*[a._to_struct() for a in _addrs])
        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRTSTADDRS, io)

        return tuple([PFTableAddr(a) for a in buffer[:io.pfrio_size] if a.pfra_fback])

    def add_addrs(self, table, *addrs):
        """Add one or more addresses to a table.

        'table' can be either a PFTable instance or a string containing the
        table name; 'addrs' can be either PFTableAddr instances or strings.
        Return the number of addresses effectively added.
        """
        if isinstance(table, str):
            table = pfr_table(pfrt_name=table.encode())
        else:
            table = pfr_table(pfrt_name=table.name.encode(),
                              pfrt_anchor=table.anchor.encode())

        _addrs = []
        for addr in addrs:
            if isinstance(addr, PFTableAddr):
                _addrs.append(addr)
            else:
                _addrs.append(PFTableAddr(addr))

        io = pfioc_table(pfrio_table=table, pfrio_esize=sizeof(pfr_addr),
                         pfrio_size=len(addrs))

        buffer = (pfr_addr * len(addrs))(*[a._to_struct() for a in _addrs])
        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRADDADDRS, io)

        return io.pfrio_nadd

    def clear_addrs(self, table):
        """Clear all addresses in the specified table.

        Return the number of addresses removed.
        """
        if isinstance(table, str):
            table = pfr_table(pfrt_name=table)
        else:
            table = pfr_table(pfrt_name=table.name.encode(),
                              pfrt_anchor=table.anchor.encode())

        io = pfioc_table(pfrio_table=table)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRCLRADDRS, io)

        return io.pfrio_ndel

    def del_addrs(self, table, *addrs):
        """Delete one or more addresses from the specified table.

        'table' can be either a PFTable instance or a string containing the
        table name; 'addrs' can be either PFTableAddr instances or strings.
        Return the number of addresses deleted.
        """
        if isinstance(table, str):
            table = pfr_table(pfrt_name=table)
        else:
            table = pfr_table(pfrt_name=table.name.encode(),
                              pfrt_anchor=table.anchor.encode())

        _addrs = []
        for addr in addrs:
            if isinstance(addr, PFTableAddr):
                _addrs.append(addr)
            else:
                _addrs.append(PFTableAddr(addr))

        io = pfioc_table(pfrio_table=table, pfrio_esize=sizeof(pfr_addr),
                         pfrio_size=len(addrs))

        buffer = (pfr_addr * len(addrs))(*[a._to_struct() for a in _addrs])
        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRDELADDRS, io)

        return io.pfrio_ndel

    def set_addrs(self, table, *addrs):
        """Replace the content of a table.

        'table' can be either a PFTable instance or a string containing the
        table name; 'addrs' can be either PFTableAddr instances or strings.
        Return a tuple containing the number of addresses deleted, added and
        changed.
        """
        if isinstance(table, str):
            table = pfr_table(pfrt_name=table)
        else:
            table = pfr_table(pfrt_name=table.name.encode(),
                              pfrt_anchor=table.anchor.encode())

        _addrs = []
        for addr in addrs:
            if isinstance(addr, PFTableAddr):
                _addrs.append(addr)
            else:
                _addrs.append(PFTableAddr(addr))

        io = pfioc_table(pfrio_table=table, pfrio_esize=sizeof(pfr_addr),
                         pfrio_size=len(addrs))

        buffer = (pfr_addr * len(addrs))(*[a._to_struct() for a in _addrs])
        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRSETADDRS, io)

        return (io.pfrio_ndel, io.pfrio_nadd, io.pfrio_nchange)

    def get_addrs(self, table, buf_size=10):
        """Get the addresses in the specified table.

        'table' can be either a PFTable instance or a string containing the
        table name. Return a list of PFTableAddr objects.
        """
        if isinstance(table, str):
            table = pfr_table(pfrt_name=table.encode())
        else:
            table = pfr_table(pfrt_name=table.name.encode(),
                              pfrt_anchor=table.anchor.encode())

        io = pfioc_table(pfrio_table=table, pfrio_esize=sizeof(pfr_addr))

        with open(self.dev, 'w') as d:
            while True:
                buffer = (pfr_addr * buf_size)()
                io.pfrio_buffer = addressof(buffer)
                io.pfrio_size = buf_size

                ioctl(d, DIOCRGETADDRS, io)

                if io.pfrio_size <= buf_size:
                    break
                buf_size = io.pfrio_size

        return tuple([PFTableAddr(a) for a in buffer[:io.pfrio_size]])

    def get_tstats(self, filter=None, buf_size=10):
        """Get statistics information for one or more tables.

        'filter' is a PFTable object that allows you to specify the anchor of
        the tables to retrieve statistics for. Return a tuple of PFTStats
        objects.
        """
        io = pfioc_table(pfrio_esize=sizeof(pfr_tstats))

        if filter is not None:
            io.pfrio_table = pfr_table(pfrt_anchor=filter.anchor)

        with open(self.dev, 'w') as d:
            while True:
                buffer = (pfr_tstats * buf_size)()
                io.pfrio_buffer = addressof(buffer)
                io.pfrio_size = buf_size

                ioctl(d, DIOCRGETTSTATS, io)

                if io.pfrio_size <= buf_size:
                    break
                buf_size = io.pfrio_size

        stats = []
        for t in buffer[:io.pfrio_size]:
            if t.pfrts_tzero:
                stats.append(PFTStats(t))

        return tuple(stats)

    def clear_tstats(self, *tables):
        """Clear the statistics of one or more tables.

        'tables' must be PFTable objects. Return the number of tables cleared.
        """
        io = pfioc_table(pfrio_esize=sizeof(pfr_table), pfrio_size=len(tables))

        buffer = (pfr_table * len(tables))()
        for (t, b) in zip(tables, buffer):
            b.pfrt_name = t.name.encode()
            b.pfrt_anchor = t.anchor.encode()

        io.pfrio_buffer = addressof(buffer)

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCRCLRTSTATS, io)

        return io.pfrio_nadd

    def get_synflood_watermarks(self):
        """Return the start and end values for adaptive syncookies watermarks"""
        ps = pfioc_synflwats()
        with open(self.dev, 'w') as d:
            ioctl(d, DIOCGETSYNFLWATS, ps)

        return (ps.hiwat, ps.lowat)

    def set_synflood_watermarks(self, start=2500, end=1500):
        """Set the start and end values for adaptive syncookies watermarks"""
        ps = pfioc_synflwats(hiwat=start, lowat=end)
        with open(self.dev, 'w') as d:
            ioctl(d, DIOCSETSYNFLWATS, ps)

    def set_syncookies(self, mode):
        """Set the syncookies mode (never, always or adaptive)"""
        if mode in pf_syncookies_modes:
            mode = pf_syncookies_modes[mode]

        with open(self.dev, 'w') as d:
            ioctl(d, DIOCSETSYNCOOKIES, c_uint8(mode))



"""Classes to represent Packet Filter Rules."""

from socket import *
from ctypes import *
import re
import pwd
import grp

from pf.exceptions import PFError
from pf.constants import *
from pf._struct import *
from pf._base import PFObject
from pf._utils import *
from pf.table import PFTable


__all__ = ['PFUid',
           'PFGid',
           'PFPort',
           'PFAddr',
           'PFRuleAddr',
           'PFPool',
           'PFDivert',
           'PFThreshold',
           'PFRule',
           'PFRuleset']


# Helper functions
def azero(seq):
    """Return True if all numbers in 'seq' are 0s."""
    return all(v == 0 for v in seq)


class PFOp(PFObject):
    """Class representing a generic comparison operation."""

    def __init__(self, num=None, op=PF_OP_NONE):
        """Check arguments and initialize instance attributes."""
        self.op = op

        if isinstance(num, str) or isinstance(num, Structure):
            super(PFOp, self).__init__(num)
        elif num is None:
            self.num = (0, 0)
        elif isinstance(num, int):
            self.num = (num, 0)
        elif isinstance(num, tuple):
            self.num = num

    def _from_struct(self, operation):
        """Initalize a new instance from a structure."""
        raise NotImplementedError

    def _from_string(self, operation):
        """Initalize a new instance from a string."""
        op_re = "(?:(?P<n1>[0-9]+)?\s*"                  + \
                   "(?P<op>=|!=|<>|><|<|<=|>|>=|:))?\s*" + \
                "(?P<n2>[0-9a-z-]+)?"

        m = re.compile(op_re).match(operation)
        if not m:
            raise ValueError("Could not parse string: {}".format(operation))

        self.op = pf_ops[m.group("op")] if m.group("op") else PF_OP_EQ

        try:
            n2 = int(m.group("n2"))
        except ValueError:
            if self.op in (PF_OP_EQ, PF_OP_NE):
                n2 = self._str_to_num(m.group("n2"))
            else:
                raise

        if self.op in (PF_OP_IRG, PF_OP_XRG, PF_OP_RRG):
            n1 = int(m.group("n1"))
            self.num = (n1, n2)
        else:
            self.num = (n2, 0)

    def _to_struct(self):
        """Return the structure representing the operation."""
        raise NotImplementedError

    def _to_string(self):
        """Return the string representation of the operation."""
        n1, n2 = self.num

        if self.op == PF_OP_NONE and not n1:
            return ""

        if self.op in (PF_OP_EQ, PF_OP_NE):
            n1 = self._num_to_str(n1)

        s = {PF_OP_NONE: "{0}",
             PF_OP_IRG:  "{0} >< {1}",
             PF_OP_XRG:  "{0} <> {1}",
             PF_OP_EQ:   "= {0}",
             PF_OP_NE:   "!= {0}",
             PF_OP_LT:   "< {0}",
             PF_OP_LE:   "<= {0}",
             PF_OP_GT:   "> {0}",
             PF_OP_GE:   ">= {0}",
             PF_OP_RRG:  "{0}:{1}"}[self.op]

        return s.format(n1, n2)

    def _num_to_str(self, n):
        """Convert a numeric operand to a string."""
        raise NotImplementedError

    def _str_to_num(self, s):
        """Convert a string to a numeric operand."""
        raise NotImplementedError

    def __eq__(self, operation):
        return (self.num == operation.num and self.op == operation.op)

    def __ne__(self, operation):
        return not self.__eq__(operation)


class PFUid(PFOp):
    """Class representing a user ID."""

    _struct_type = pf_rule_uid

    def __init__(self, num=None, op=PF_OP_NONE):
        """Check arguments and initialize instance attributes."""
        super(PFUid, self).__init__(num, op)

    def _from_struct(self, uid):
        """Initialize a new instance from a pf_rule_uid structure."""
        self.num = tuple(uid.uid)
        self.op  = uid.op

    def _to_struct(self):
        """Convert this instance to a pf_rule_uid structure."""
        return pf_rule_uid(self.num, self.op)

    def _num_to_str(self, n):
        """Convert a numeric user ID to a string."""
        try:
            return pwd.getpwuid(n).pw_name
        except KeyError:
            return n

    def _str_to_num(self, s):
        """Convert a string to a numeric user ID."""
        return pwd.getpwnam(s).pw_uid


class PFGid(PFOp):
    """Class representing a group ID."""

    _struct_type = pf_rule_gid

    def __init__(self, num=None, op=PF_OP_NONE):
        """Check arguments and initialize instance attributes."""
        super(PFGid, self).__init__(num, op)

    def _from_struct(self, gid):
        """Initialize a new instance from a pf_rule_gid structure."""
        self.num = tuple(gid.gid)
        self.op  = gid.op

    def _to_struct(self):
        """Convert this instance to a pf_rule_gid structure."""
        return pf_rule_gid(self.num, self.op)

    def _num_to_str(self, n):
        """Convert a numeric group ID to a string."""
        try:
            return grp.getgrgid(n).gr_name
        except KeyError:
            return n

    def _str_to_num(self, s):
        """Convert a string to a numeric group ID."""
        return grp.getgrnam(s).gr_gid


class PFPort(PFOp):
    """Class representing a TCP/UDP port."""

    def __init__(self, num=None, proto=None, op=PF_OP_NONE):
        """Check arguments and initialize instance attributes."""
        self.proto = proto
        super(PFPort, self).__init__(num, op)

    def _num_to_str(self, n):
        """Convert a numeric port to a service name."""
        return n
        #try:
        #    return getservbyport(n, getprotobynumber(self.proto))
        #except (TypeError, error):
        #    return n

    def _str_to_num(self, s):
        """Convert a service name to a numeric port."""
        return getservbyname(s, getprotobynumber(self.proto))

    def __eq__(self, p):
        return (self.num   == p.num and
                self.op    == p.op  and
                self.proto == p.proto)


class PFAddr(PFObject):
    """Class representing an address."""

    _struct_type = pf_addr_wrap

    def __init__(self, addr=None, af=AF_UNSPEC, **kw):
        """Check arguments and initialize instance attributes."""
        self.af = af

        if addr is None:
            t = kw.get("type", PF_ADDR_ADDRMASK)
            addr = pf_addr_wrap(type=t)

        super(PFAddr, self).__init__(addr, **kw)

    def _from_struct(self, a):
        """Initalize a new instance from a pf_addr_wrap structure."""
        self.type = a.type
        a6, m6 = a.v.a.addr.v6, a.v.a.mask.v6

        if self.type == PF_ADDR_DYNIFTL  and self.af != AF_UNSPEC          or \
           self.type == PF_ADDR_ADDRMASK and not (azero(a6) and azero(m6)) or \
           self.type == PF_ADDR_RANGE:
            try:
                l = {AF_INET: 4, AF_INET6: 16}[self.af]
            except KeyError:
                raise PFError("No valid address family specified")
            else:
                addr = inet_ntop(self.af, string_at(addressof(a6), l))
                mask = inet_ntop(self.af, string_at(addressof(m6), l))
        else:
            addr = mask = None

        if self.type == PF_ADDR_DYNIFTL:
            self.ifname = a.v.ifname.decode()
            self.iflags = a.iflags
            self.dyncnt = a.p.dyncnt
            self.mask = mask
        elif self.type == PF_ADDR_TABLE:
            self.tblname = a.v.tblname.decode()
            self.tblcnt = a.p.tblcnt
        elif self.type == PF_ADDR_ADDRMASK:
            self.addr = addr
            self.mask = mask
        elif self.type == PF_ADDR_RTLABEL:
            self.rtlabelname = a.v.rtlabelname.decode()
            self.rtlabel = a.v.rtlabel.decode()
        elif self.type == PF_ADDR_RANGE:
            self.addr = (addr, mask)

    def _from_string(self, a):
        """Initalize a new instance from a string."""
        ipv4_re = "[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"
        ipv6_re = "[0-9a-f:]+"
        addr_re = "(?P<nort>no-route)|"                                 + \
                  "(?P<urpf>urpf-failed)|"                              + \
                  "(?P<any>any)|"                                       + \
                  "(?P<tbl><(?P<tblname>\w+)>)|"                        + \
                  "(?P<rt>route\s+(?P<rtlbl>\w+))|"                     + \
                  "(?P<if>\((?P<ifname>[a-z]+[0-9]*)"                   + \
                           "(?P<mod>(:network|:broadcast|:peer|:0)*)\)" + \
                           "(?:/(?P<ifmask>\d+))?)|"                    + \
                  "(?P<ipv4rg>(?P<ipv4_1>{})\s*-\s*".format(ipv4_re)    + \
                             "(?P<ipv4_2>{}))|".format(ipv4_re)         + \
                  "(?P<ipv6rg>(?P<ipv6_1>{})\s*-\s*".format(ipv6_re)    + \
                             "(?P<ipv6_2>{}))|".format(ipv6_re)         + \
                  "(?P<ipv4>{})(?:/(?P<mask4>\d+))?|".format(ipv4_re)   + \
                  "(?P<ipv6>{})(?:/(?P<mask6>\d+))?".format(ipv6_re)

        m = re.compile(addr_re).match(a)
        if not m:
            raise ValueError("Could not parse address: {}".format(a))

        if m.group("nort"):
            self.type = PF_ADDR_NOROUTE
        elif m.group("urpf"):
            self.type = PF_ADDR_URPFFAILED
        elif m.group("any"):
            self.type = PF_ADDR_ADDRMASK
            self.addr = self.mask = None
        elif m.group("tbl"):
            self.type = PF_ADDR_TABLE
            self.tblname = m.group("tblname")
            self.tblcnt = -1
        elif m.group("rt"):
            self.type = PF_ADDR_RTLABEL
            self.rtlabelname = m.group("rtlbl")
            self.rtlabel = 0
        elif m.group("if"):
            self.type = PF_ADDR_DYNIFTL
            self.ifname = m.group("ifname")
            try:
                b = {AF_INET: 32, AF_INET6: 128}[self.af]
            except KeyError:
                self.mask = None
            else:
                if (m.group("ifmask")):
                    self.mask = ctonm(int(m.group("ifmask")), self.af)
                else:
                    self.mask = ctonm(b, self.af)
            self.dyncnt = 0
            self.iflags = 0
            for mod in m.group("mod").split(":")[1:]:
                self.iflags |= pf_if_mods[mod]
        elif m.group("ipv4rg"):
            self.af = AF_INET
            self.type = PF_ADDR_RANGE
            self.addr = m.group("ipv4_1", "ipv4_2")
        elif m.group("ipv6rg"):
            self.af = AF_INET6
            self.type = PF_ADDR_RANGE
            self.addr = m.group("ipv6_1", "ipv6_2")
        elif m.group("ipv4"):
            self.af = AF_INET
            self.type = PF_ADDR_ADDRMASK
            self.addr = m.group("ipv4")
            if m.group("mask4"):
                self.mask = ctonm(int(m.group("mask4")), self.af)
            else:
                self.mask = ctonm(32, self.af)
        elif m.group("ipv6"):
            self.af = AF_INET6
            self.type = PF_ADDR_ADDRMASK
            self.addr = m.group("ipv6")
            if m.group("mask6"):
                self.mask = ctonm(int(m.group("mask6")), self.af)
            else:
                self.mask = ctonm(128, self.af)

    def _to_struct(self):
        """Convert this instance to a pf_addr_wrap structure."""
        a = pf_addr_wrap()
        a.type = self.type

        if self.type == PF_ADDR_DYNIFTL:
            a.v.ifname = self.ifname.encode()
            a.p.dyncnt = self.dyncnt
            a.iflags = self.iflags
            if self.af == AF_UNSPEC:
                mask = b'\xff' * 16
            else:
                mask = inet_pton(self.af, self.mask)
            memmove(a.v.a.mask.v6, c_char_p(mask), len(mask))
        elif self.type == PF_ADDR_TABLE:
            a.v.tblname = self.tblname.encode()
            a.p.tblcnt = self.tblcnt
        elif self.type == PF_ADDR_ADDRMASK and self.addr:
            addr = inet_pton(self.af, self.addr)
            mask = inet_pton(self.af, self.mask)
            memmove(a.v.a.addr.v6, c_char_p(addr), len(addr))
            memmove(a.v.a.mask.v6, c_char_p(mask), len(mask))
        elif self.type == PF_ADDR_RTLABEL:
            a.v.rtlabelname = self.rtlabelname.encode()
            a.v.rtlabel = self.rtlabel
        elif self.type == PF_ADDR_RANGE:
            addr1 = inet_pton(self.af, self.addr[0])
            addr2 = inet_pton(self.af, self.addr[1])
            memmove(a.v.a.addr.v6, c_char_p(addr1), len(addr1))
            memmove(a.v.a.mask.v6, c_char_p(addr2), len(addr2))

        return a

    def _to_string(self):
        """Return the string representation of the address."""
        if self.type == PF_ADDR_DYNIFTL:
            s = "({.ifname}".format(self)
            if self.iflags & PFI_AFLAG_NETWORK:
                s += ":network"
            if self.iflags & PFI_AFLAG_BROADCAST:
                s += ":broadcast"
            if self.iflags & PFI_AFLAG_PEER:
                s += ":peer"
            if self.iflags & PFI_AFLAG_NOALIAS:
                s += ":0"
            s += ")"
        elif self.type == PF_ADDR_TABLE:
            return "<{.tblname}>".format(self)
        elif self.type == PF_ADDR_ADDRMASK:
            s = self.addr or "any"
        elif self.type == PF_ADDR_NOROUTE:
            return "no-route"
        elif self.type == PF_ADDR_URPFFAILED:
            return "urpf-failed"
        elif self.type == PF_ADDR_RTLABEL:
            return "route \"{.rtlabelname}\"".format(self)
        elif self.type == PF_ADDR_RANGE:
            s = "{0.addr[0]} - {0.addr[1]}".format(self)
        else:
            return "?"

        if self.type != PF_ADDR_RANGE and self.mask:
            bits = nmtoc(self.mask, self.af)
            if not ((self.af == AF_INET and bits == 32) or (bits == 128)):
                s += "/{}".format(bits)

        return s

    def _is_any(self):
        """Return true if this address matches any host."""
        return (self.type == PF_ADDR_ADDRMASK and self.addr is None)

    def __eq__(self, a):
        if (self.type != a.type or self.af != a.af):
            return False

        if self.type == PF_ADDR_DYNIFTL:
            return (self.ifname == a.ifname and
                    self.mask   == a.mask   and
                    self.iflags == a.iflags)
        elif self.type == PF_ADDR_TABLE:
            return (self.tblname == a.tblname)
        elif self.type == PF_ADDR_ADDRMASK:
            return (self.addr == a.addr and
                    self.mask == a.mask)
        elif self.type == PF_ADDR_RTLABEL:
            return (self.rtlabelname == a.rtlabelname.decode())
        elif self.type == PF_ADDR_RANGE:
            return (self.addr == a.addr)

        return True

    def __ne__(self, a):
        return not self.__eq__(a)


class PFRuleAddr(PFObject):
    """Class representing an address/port pair."""

    _struct_type = pf_rule_addr

    def __init__(self, addr=None, port=None, neg=False, **kw):
        """Check arguments and initialize instance attributes."""
        if isinstance(addr, self._struct_type):
            self.addr   = PFAddr(addr.addr, kw['af'])
            self.port   = PFPort(tuple(map(ntohs, addr.port)),
                                 kw['proto'], addr.port_op)
            self.neg    = bool(addr.neg)
            self.weight = addr.weight
        else:
            self.addr   = addr or PFAddr()
            self.port   = port or PFPort()
            self.neg    = bool(neg)
            self.weight = 0

    def _to_struct(self):
        """Convert this instance to a pf_rule_addr structure."""
        a = pf_rule_addr()

        a.addr    = self.addr._to_struct()
        a.port    = tuple(map(htons, self.port.num))
        a.port_op = self.port.op
        a.neg     = int(self.neg)
        a.weight  = self.weight

        return a

    def _to_string(self):
        """Return the string representation of the address/port pair."""
        s = ("! {.addr}" if self.neg else "{.addr}").format(self)
        p = "{.port}".format(self)
        if p:
            s += (":" if self.port.op == PF_OP_NONE else " port ") + p

        return s

    def __eq__(self, a):
        return (self.addr == a.addr and
                self.port == a.port and
                self.neg  == a.neg)

    def __ne__(self, a):
        return not self.__eq__(a)


class PFPool(PFObject):
    """Class representing an address pool."""

    _struct_type = pf_pool

    def __init__(self, id, pool, **kw):
        """Check arguments and initialize instance attributes."""
        self.id = id

        if isinstance(pool, PFAddr):
            self._af = pool.af
            p = pf_pool(addr=pool._to_struct())
            if self.id == PF_POOL_NAT:
                p.proxy_port = (PF_NAT_PROXY_PORT_LOW, PF_NAT_PROXY_PORT_HIGH)
        elif isinstance(pool, self._struct_type):
            self._af = kw.pop("af", AF_UNSPEC)
            p = pool

        super(PFPool, self).__init__(p, **kw)

    def _from_struct(self, p):
        """Initalize a new instance from a pf_pool structure."""
        self.addr       = PFAddr(p.addr, self._af)
        self.key        = "{:#010x}{:08x}{:08x}{:08x}".format(*p.key.key32)
        self.counter    = p.counter
        self.ifname     = p.ifname.decode()
        self.tblidx     = p.tblidx
        self.states     = p.states
        self.curweight  = p.curweight
        self.weight     = p.weight
        self.proxy_port = PFPort(tuple(p.proxy_port), op=p.port_op)
        self.opts       = p.opts

    def _to_struct(self):
        """Convert a PFPool object to a pf_pool structure."""
        p = pf_pool()

        if self.addr._is_any():
            p.addr = PFAddr(type=PF_ADDR_NONE)._to_struct()
        else:
            p.addr       = self.addr._to_struct()
        p.ifname     = self.ifname.encode()
        p.states     = self.states
        p.curweight  = self.curweight
        p.weight     = self.weight
        p.proxy_port = self.proxy_port.num
        p.port_op    = self.proxy_port.op
        p.opts       = self.opts

        return p

    def _to_string(self):
        """Return the string representation of the address pool."""
        p1, p2 = self.proxy_port.num
        s = ""

        if self.ifname:
            if self.addr.addr is not None:
                s += "{.addr}@".format(self)
            s += self.ifname
        else:
            s += "{.addr}".format(self)

        if self.id == PF_POOL_NAT:
            if (p1, p2) != (PF_NAT_PROXY_PORT_LOW, PF_NAT_PROXY_PORT_HIGH) and \
               (p1, p2) != (0, 0):
                if p1 == p2:
                    s += " port {}".format(p1)
                else:
                    s += " port {}:{}".format(p1, p2)
        elif self.id == PF_POOL_RDR:
            if p1:
                s += " port {}".format(p1)
                if p2 and (p2 != p1):
                    s += ":{}".format(p2)

        opt = self.opts & PF_POOL_TYPEMASK
        if opt == PF_POOL_BITMASK:
            s += " bitmask"
        elif opt == PF_POOL_RANDOM:
            s += " random"
        elif opt == PF_POOL_SRCHASH:
            s += " source-hash {0.key}".format(self)
        elif opt == PF_POOL_ROUNDROBIN:
            s += " round-robin"
        elif opt == PF_POOL_LEASTSTATES:
            s += " least-states"

        if self.opts & PF_POOL_STICKYADDR:
            s += " sticky-address"

        if (self.id == PF_POOL_NAT) and (p1 == p2 == 0):
            s += " static-port"

        return s


class PFDivert(PFObject):
    """Class representing a divert socket."""

    _struct_type = divert

    def __init__(self, type=PF_DIVERT_NONE, addr=None, port=None, af=AF_UNSPEC):
        """Check arguments and initialize instance attributes."""
        self.af = af
        if isinstance(type, self._struct_type):
            self._from_struct(type)
        else:
            self.type = type
            self.addr = addr
            self.port = port
            if addr and (af == AF_UNSPEC):
                if is_IPaddr(addr):
                    self.af = AF_INET
                elif is_IP6addr(addr):
                    self.af = AF_INET6

    def _from_struct(self, d):
        """ Initalize a new instance from a divert structure."""
        self.type = d.type
        self.addr = None
        self.port = None

        if self.type == PF_DIVERT_TO:
            self.addr = inet_ntop(self.af, string_at(addressof(d.addr), l))
            self.port = ntohs(d.port)
        elif self.type == PF_DIVERT_PACKET:
            self.port = ntohs(d.port)

    def _to_struct(self):
        """Convert a PFDivert object to a divert structure."""
        d = divert(type=self.type)
        if self.type in (PF_DIVERT_TO, PF_DIVERT_PACKET):
            d.port = htons(self.port)
        if self.type == PF_DIVERT_TO:
            d.addr = inet_pton(self.af, self.addr)

        return d

    def _to_string(self):
        """Return a string representation of the object."""
        if self.type == PF_DIVERT_NONE:
            return ""
        if self.type == PF_DIVERT_TO:
            return "divert-to {.addr} port {.port}".format(self)
        elif self.type == PF_DIVERT_REPLY:
            return "divert-reply"
        elif self.type == PF_DIVERT_PACKET:
            return "divert-packet port {.port}".format(self)
        else:
            return "divert ???"


class PFThreshold(PFObject):
    """Measure the rate of packets matching the rule and states created by it"""

    _struct_type = pf_threshold

    def __init__(self, limit, seconds=0):
        """Check arguments and initialize instance attributes"""
        if not isinstance(limit, self._struct_type):
            limit = pf_threshold(limit=limit, seconds=seconds)
        self._from_struct(limit)

    def _from_struct(self, threshold):
        """Initalize a new instance from a pf_threshold structure"""
        self.limit   = threshold.limit
        self.seconds = threshold.seconds
        self.count   = threshold.count
        self.last    = threshold.last

    def _to_struct(self):
        """Convert a PFThreshold object to a pf_threshold structure."""
        threshold = pf_threshold()

        threshold.limit   = self.limit
        threshold.seconds = self.seconds
        threshold.count   = self.count
        threshold.last    = self.last

        return threshold

    def _to_string(self):
        """Return a string representation of the object."""
        return "max-pkt-rate {.limit}/{.seconds}".format(self)


class PFRule(PFObject):
    """Class representing a Packet Filter rule."""

    _struct_type = pf_rule

    def __init__(self, rule=None, **kw):
        """Check arguments and initialize instance attributes."""
        if rule is None:
            rule = pf_rule(rtableid=-1, onrdomain=-1)
        super(PFRule, self).__init__(rule, **kw)

    def _from_struct(self, r):
        """Initalize a new instance from a pf_rule structure."""
        self.src               = PFRuleAddr(r.src, af=r.af, proto=r.proto)
        self.dst               = PFRuleAddr(r.dst, af=r.af, proto=r.proto)
        #skip
        self.label             = r.label
        self.ifname            = r.ifname.decode()
        self.rcv_ifname        = r.rcv_ifname.decode()
        self.qname             = r.qname
        self.pqname            = r.pqname
        self.tagname           = r.tagname
        self.match_tagname     = r.match_tagname
        self.overload_tblname  = r.overload_tblname
        #entries

        self.nat               = PFPool(PF_POOL_NAT, r.nat, af=r.af)
        if self.nat.addr._is_any():
            self.nat = PFPool(PF_POOL_NAT, PFAddr(type=PF_ADDR_NONE))
        self.rdr               = PFPool(PF_POOL_RDR, r.rdr, af=r.af)
        if self.rdr.addr._is_any():
            self.rdr = PFPool(PF_POOL_RDR, PFAddr(type=PF_ADDR_NONE))
        self.route             = PFPool(PF_POOL_ROUTE, r.route, af=r.af)
        if self.route.addr._is_any():
            self.route = PFPool(PF_POOL_ROUTE, PFAddr(type=PF_ADDR_NONE))

        self.pktrate           = PFThreshold(r.pktrate)
        self.evaluations       = r.evaluations
        self.packets           = tuple(r.packets)
        self.bytes             = tuple(r.bytes)
        self.os_fingerprint    = r.os_fingerprint
        self.rtableid          = r.rtableid
        self.onrdomain         = r.onrdomain
        self.timeout           = list(r.timeout)
        self.states_cur        = r.states_cur
        self.states_tot        = r.states_tot
        self.max_states        = r.max_states
        self.src_nodes         = r.src_nodes
        self.max_src_nodes     = r.max_src_nodes
        self.max_src_states    = r.max_src_states
        self.max_src_conn      = r.max_src_conn
        self.max_src_conn_rate = (r.max_src_conn_rate.limit,
                                  r.max_src_conn_rate.seconds)
        self.qid               = r.qid
        self.pqid              = r.pqid
        self.rt_listid         = r.rt_listid
        self.nr                = r.nr
        self.prob              = r.prob
        self.cuid              = r.cuid
        self.cpid              = r.cpid
        self.return_icmp       = r.return_icmp
        self.return_icmp6      = r.return_icmp6
        self.max_mss           = r.max_mss
        self.tag               = r.tag
        self.match_tag         = r.match_tag
        self.scrub_flags       = r.scrub_flags
        self.delay             = r.delay
        self.uid               = PFUid(r.uid)
        self.gid               = PFGid(r.gid)
        self.rule_flag         = r.rule_flag
        self.action            = r.action
        self.direction         = r.direction
        self.log               = r.log
        self.logif             = r.logif
        self.quick             = bool(r.quick)
        self.ifnot             = bool(r.ifnot)
        self.match_tag_not     = bool(r.match_tag_not)
        self.keep_state        = r.keep_state
        self.af                = r.af
        self.proto             = r.proto
        self.type              = r.type
        self.code              = r.code
        self.flags             = "".join([f for n, f in enumerate("FSRPAUEW")
                                            if r.flags & (1 << n)])
        self.flagset           = "".join([f for n, f in enumerate("FSRPAUEW")
                                            if r.flagset & (1 << n)])
        self.min_ttl           = r.min_ttl
        self.allow_opts        = bool(r.allow_opts)
        self.rt                = r.rt
        self.return_ttl        = r.return_ttl
        self.tos               = r.tos
        self.set_tos           = r.set_tos
        self.anchor_relative   = r.anchor_relative
        self.anchor_wildcard   = r.anchor_wildcard
        self.flush             = r.flush
        self.prio              = r.prio
        self.set_prio          = tuple(r.set_prio)
        self.naf               = r.naf
        self.divert            = PFDivert(r.divert, af=self.af)

    def _to_struct(self):
        """Convert a PFRule object to a pf_rule structure."""
        r = pf_rule()

        r.src               = self.src._to_struct()
        r.dst               = self.dst._to_struct()
        #skip
        r.label             = self.label
        r.ifname            = self.ifname.encode()
        r.rcv_ifname        = self.rcv_ifname.encode()
        r.qname             = self.qname
        r.pqname            = self.pqname
        r.tagname           = self.tagname
        r.match_tagname     = self.match_tagname
        r.overload_tblname  = self.overload_tblname
        #entries
        r.nat               = self.nat._to_struct()
        r.rdr               = self.rdr._to_struct()
        r.route             = self.route._to_struct()
        r.pktrate           = self.pktrate._to_struct()
        r.evaluations       = self.evaluations
        r.packets           = self.packets
        r.bytes             = self.bytes
        r.os_fingerprint    = self.os_fingerprint
        r.rtableid          = self.rtableid
        r.onrdomain         = self.onrdomain
        for i, t in enumerate(self.timeout):
            r.timeout[i] = t
        r.states_cur        = self.states_cur
        r.states_tot        = self.states_tot
        r.max_states        = self.max_states
        r.src_nodes         = self.src_nodes
        r.max_src_nodes     = self.max_src_nodes
        r.max_src_states    = self.max_src_states
        r.max_src_conn      = self.max_src_conn
        r.max_src_conn_rate = self.max_src_conn_rate
        r.qid               = self.qid
        r.pqid              = self.pqid
        r.rt_listid         = self.rt_listid
        r.nr                = self.nr
        r.prob              = self.prob
        r.cuid              = self.cuid
        r.cpid              = self.cpid
        r.return_icmp       = self.return_icmp
        r.return_icmp6      = self.return_icmp6
        r.max_mss           = self.max_mss
        r.tag               = self.tag
        r.match_tag         = self.match_tag
        r.scrub_flags       = self.scrub_flags
        r.delay             = self.delay
        r.uid               = self.uid._to_struct()
        r.gid               = self.gid._to_struct()
        r.rule_flag         = self.rule_flag
        r.action            = self.action
        r.direction         = self.direction
        r.log               = self.log
        r.logif             = self.logif
        r.quick             = int(self.quick)
        r.ifnot             = int(self.ifnot)
        r.match_tag_not     = int(self.match_tag_not)
        r.keep_state        = self.keep_state
        r.af                = self.af
        r.proto             = self.proto
        r.type              = self.type
        r.code              = self.code
        r.flags             = sum([1<<"FSRPAUEW".find(f) for f in self.flags])
        r.flagset           = sum([1<<"FSRPAUEW".find(f) for f in self.flagset])
        r.min_ttl           = self.min_ttl
        r.allow_opts        = int(self.allow_opts)
        r.rt                = self.rt
        r.return_ttl        = self.return_ttl
        r.tos               = self.tos
        r.set_tos           = self.set_tos
        r.anchor_relative   = self.anchor_relative
        r.anchor_wildcard   = self.anchor_wildcard
        r.flush             = self.flush
        r.prio              = self.prio
        r.set_prio[:]       = self.set_prio
        r.naf               = self.naf
        r.divert            = self.divert._to_struct()

        return r

    def _to_string(self):
        """Return the string representation of the rule."""
        pf_actions = ("pass", "block", "scrub", "no scrub", "nat", "no nat",
                      "binat", "no binat", "rdr", "no rdr", "", "", "match")
        pf_anchors = ("anchor", "anchor", "anchor", "anchor", "nat-anchor",
                      "nat-anchor", "binat-anchor", "binat-anchor",
                      "rdr-anchor", "rdr-anchor")

        if self.rule_flag & PFRULE_EXPIRED:
            return ""

        if self.action > PF_MATCH:
            s = "action({.action})".format(self)
        elif isinstance(self, PFRuleset):
            s = pf_anchors[self.action]
            if not self.name.startswith("_"):
                s += " \"{.name}\"".format(self)
        else:
            s = pf_actions[self.action]

        if self.action == PF_DROP:
            if self.rule_flag & PFRULE_RETURN:
                s += " return"
            elif self.rule_flag & PFRULE_RETURNRST:
                s += " return-rst"
                if self.return_ttl:
                    s += "(ttl {.return_ttl})".format(self)
            elif self.rule_flag & PFRULE_RETURNICMP:
                ic  = geticmpcodebynumber(self.return_icmp >> 8,
                                          self.return_icmp & 0xff, AF_INET)
                ic6 = geticmpcodebynumber(self.return_icmp6 >> 8,
                                          self.return_icmp6 & 0xff, AF_INET6)
                s += " return-icmp"
                if self.af == AF_INET:
                    s += "({})".format(ic or self.return_icmp & 0xff)
                elif self.af == AF_INET6:
                    s += "6({})".format(ic6 or self.return_icmp6 & 0xff)
                else:
                    s += "({}, {})".format((ic or self.return_icmp & 0xff),
                                           (ic6 or self.return_icmp6 & 0xff))
            else:
                s += " drop"

        if self.direction == PF_IN:
            s += " in"
        elif self.direction == PF_OUT:
            s += " out"

        if self.log:
            s += " log"
            if (self.log & ~PF_LOG) or self.logif:
                l = []
                if self.log & PF_LOG_ALL:
                    l.append("all")
                if self.log & PF_LOG_MATCHES:
                    l.append("matches")
                if self.log & PF_LOG_SOCKET_LOOKUP:
                    l.append("user")
                if self.logif:
                    l.append("to pflog{.logif}".format(self))
                s += " ({})".format(", ".join(l))

        if self.quick:
            s += " quick"

        if self.ifname and self.ifname != "all":
            # "on all" not printed because it would make
            # the rule not parseable by pfctl
            if self.ifnot:
                s += " on ! {.ifname}".format(self)
            else:
                s += " on {.ifname}".format(self)

        if self.onrdomain >= 0:
            if self.ifnot:
                s += " on ! rdomain {.onrdomain}".format(self)
            else:
                s += " on rdomain {.onrdomain}".format(self)

        if self.af:
            s += " inet" if (self.af == AF_INET) else " inet6"

        if self.proto:
            s += " proto {}".format(getprotobynumber(self.proto) or self.proto)

        if self.src.addr._is_any() and self.dst.addr._is_any() and \
           not self.src.neg and not self.dst.neg               and \
           not self.src.port.op and not self.dst.port.op       and \
           self.os_fingerprint == PF_OSFP_ANY:
            s += " all"
        else:
            s += " from {.src}".format(self)
            #if self.os_fingerprint != PF_OSFP_ANY:
            s += " to {.dst}".format(self)

        if self.rcv_ifname:
            s += " received on {.rcv_ifname}".format(self)
        if self.uid.op:
            s += " user {.uid}".format(self)
        if self.gid.op:
            s += " group {.gid}".format(self)

        if self.flags or self.flagset:
            s += " flags {0.flags}/{0.flagset}".format(self)
        elif self.action in (PF_PASS, PF_MATCH)     and \
             self.proto in (0, IPPROTO_TCP)         and \
             not (self.rule_flag & PFRULE_FRAGMENT) and \
             not isinstance(self, PFRuleset)        and \
             self.keep_state:
            s += " flags any"

        if self.type:
            it = geticmptypebynumber(self.type-1, self.af)
            if self.af != AF_INET6:
                s += " icmp-type"
            else:
                s += " icmp6-type"
            s += " {}".format(it or self.type-1)
            if self.code:
                ic = geticmpcodebynumber(self.type-1, self.code-1, self.af)
                s += " code {}".format(ic or self.code-1)

        if self.tos:
            s += " tos {.tos:#04x}".format(self)
        if self.prio:
            prio = 0 if (self.prio == PF_PRIO_ZERO) else self.prio
            s += " prio {}".format(prio)
        if self.pktrate.limit:
            s += " {.pktrate}".format(self)

        if self.scrub_flags & PFSTATE_SETMASK or self.qname or \
           self.rule_flag & PFRULE_SETDELAY:
            opts = []
            if self.scrub_flags[0] != PFSTATE_SETPRIO:
                if self.set_prio[0] == self.set_prio[1]:
                    opts.append("prio {}".format(self.set_prio[0]))
                else:
                    opts.append("prio ({}, {})".format(*self.set_prio))
            if self.qname:
                if self.pqname:
                    opts.append("queue ({.qname}, {.pqname})".format(self))
                else:
                    opts.append("queue {.qname}".format(self))
            if self.scrub_flags & PFSTATE_SETTOS:
                opts.append("tos {.set_tos:#04x}".format(self))
            if self.rule_flag & PFRULE_SETDELAY:
                opts.append("delay {.delay}".format(self))
            s += " set ( {} )".format(", ".join(opts))

        has_opts = False
        if (self.max_states or self.max_src_nodes or self.max_src_states)  or \
           self.rule_flag & (PFRULE_NOSYNC|PFRULE_SRCTRACK|PFRULE_IFBOUND) or \
           self.rule_flag & (PFRULE_STATESLOPPY|PFRULE_PFLOW)              or \
           any(self.timeout):
            has_opts = True

        if not self.keep_state and self.action == PF_PASS and \
           not isinstance(self, PFRuleset):
            s += " no state"
        elif (self.keep_state == PF_STATE_NORMAL) and has_opts:
            s += " keep state"
        elif self.keep_state == PF_STATE_MODULATE:
            s += " modulate state"
        elif self.keep_state == PF_STATE_SYNPROXY:
            s += " synproxy state"

        if self.prob:
            s += " probability {:.0f}%".format(self.prob*100.0/(UINT_MAX+1))

        if has_opts:
            opts = []
            if self.max_states:
                opts.append("max {.max_states}".format(self))
            if self.rule_flag & PFRULE_NOSYNC:
                opts.append("no-sync")
            if self.rule_flag & PFRULE_SRCTRACK:
                if self.rule_flag & PFRULE_RULESRCTRACK:
                    opts.append("source-track rule")
                else:
                    opts.append("source-track global")
            if self.max_src_states:
                opts.append("max-src-states {.max_src_states}".format(self))
            if self.max_src_conn:
                opts.append("max-src-conn {.max_src_conn}".format(self))
            if self.max_src_conn_rate[0]:
                opts.append("max-src-conn-rate " +
                            "{}/{}".format(*self.max_src_conn_rate))
            if self.max_src_nodes:
                opts.append("max-src-nodes {.max_src_nodes}".format(self))
            if self.overload_tblname:
                opt = "overload <{.overload_tblname}>".format(self)
                if self.flush:
                    opt += " flush"
                if self.flush & PF_FLUSH_GLOBAL:
                    opt += " global"
                opts.append(opt)
            if self.rule_flag & PFRULE_IFBOUND:
                opts.append("if-bound")
            if self.rule_flag & PFRULE_STATESLOPPY:
                opts.append("sloppy")
            if self.rule_flag & PFRULE_PFLOW:
                opts.append("pflow")
            for i, t in enumerate(self.timeout):
                if t:
                    tm = [k for (k, v) in pf_timeouts.items() if v == i][0]
                    opts.append("{} {}".format(tm, t))

            s += " ({})".format(", ".join(opts))

        if self.rule_flag & PFRULE_FRAGMENT:
            s += " fragment"

        if self.scrub_flags >= PFSTATE_NODF or self.min_ttl or self.max_mss:
            opts = []
            if self.scrub_flags & PFSTATE_NODF:
                opts.append("no-df")
            if self.scrub_flags & PFSTATE_RANDOMID:
                opts.append("random-id")
            if self.min_ttl:
                opts.append("min-ttl {.min_ttl}".format(self))
            if self.scrub_flags & PFSTATE_SCRUB_TCP:
                opts.append("reassemble tcp")
            if self.max_mss:
                opts.append("max_mss {.max_mss}".format(self))
            s += " scrub ({})".format(" ".join(opts))

        if self.allow_opts:
            s += " allow-opts"
        if self.label:
            s += " label \"{.label}\"".format(self)
        if self.rule_flag & PFRULE_ONCE:
            s += " once"

        if self.qname and self.pqname:
            s += " queue({0.qname}, {0.pqname})".format(self)
        elif self.qname:
            s += " queue {.qname}".format(self)

        if self.tagname:
            s += " tag {.tagname}".format(self)
        if self.match_tagname:
            if self.match_tag_not:
                s += " !"
            s += " tagged {.match_tagname}".format(self)

        if self.rtableid != -1:
            s += " rtable {.rtableid}".format(self)

        if self.divert.type != PF_DIVERT_NONE:
            s += " {.divert}".format(self)

        if not isinstance(self, PFRuleset):
            if self.nat.addr.type != PF_ADDR_NONE:
                if self.rule_flag & PFRULE_AFTO:
                    af = " inet" if (self.naf == AF_INET) else " inet6"
                    s += " af-to {} from {.nat}".format(af, self)
                    if self.rdr.addr.type != PF_ADDR_NONE:
                        s += " to {.rdr}".format(self)
                else:
                    s += " nat-to {.nat}".format(self)
            elif self.rdr.addr.type != PF_ADDR_NONE:
                s += " rdr-to {.rdr}".format(self)

        if self.rt == PF_ROUTETO:
            s += " route-to {.route}".format(self)
        elif self.rt == PF_REPLYTO:
            s += " reply-to {.route}".format(self)
        elif self.rt == PF_DUPTO:
            s += " dup-to {.route}".format(self)

        return s


class PFRuleset(PFRule):
    """Class representing a Packet Filter ruleset or anchor."""

    def __init__(self, name="", rule=None, **kw):
        """Check arguments and initialize instance attributes."""
        self.name    = name
        self._altqs  = []
        self._tables = []
        self._rules  = []
        super(PFRuleset, self).__init__(rule, **kw)

    def append(self, *items):
        """Append one or more rules and/or tables and/or altqs."""
        self._rules.extend(filter(lambda i: isinstance(i, PFRule), items))
        self._tables.extend(filter(lambda i: isinstance(i, PFTable), items))

    def insert(self, index, rule):
        """Insert a 'rule' before 'index'."""
        self._rules.insert(index, rule)

    def remove(self, index=-1):
        """Remove the rule at 'index'."""
        self._rules.pop(index)

    @property
    def rules(self):
        """Return the rules in this ruleset."""
        return self._rules

    @property
    def tables(self):
        """Return the tables in this ruleset."""
        return self._tables

    def _to_string(self):
        """Return the string representation of the ruleset."""
        return "\n".join([PFRule._to_string(rule) for rule in self._rules])



"""Command for running unit tests for the module from setup.py"""

from distutils.cmd import Command
from unittest import TextTestRunner, TestLoader

import pf.tests


class TestCommand(Command):
    """Command for running unit tests for the module from setup.py"""

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        suite = TestLoader().loadTestsFromModule(pf.tests)
        TextTestRunner(verbosity=1).run(suite)



"""Classes representing the internal Packet Filter statistics and counters.

PFStatus objects contain a series of runtime statistical information describing
the current status of the Packet Filter.
"""

import time
from socket import ntohl

from pf.constants import *
from pf._struct import pf_status, pfi_kif
from pf._base import PFObject
from pf._utils import dbg_levels, uptime


__all__ = ['PFStatus',
           'PFIface']


class PFStatus(PFObject):
    """Class representing the internal Packet Filter statistics and counters."""

    _struct_type = pf_status

    def __init__(self, status):
        """Check argument and initialize class attributes."""
        super(PFStatus, self).__init__(status)

    def _from_struct(self, s):
        """Initialize class attributes from a pf_status structure."""
        self.ifname          = s.ifname.decode()
        self.running         = bool(s.running)
        self.stateid         = s.stateid
        self.syncookies_inflight = s.syncookies_inflight
        self.syncookies_active = s.syncookies_active
        self.syncookies_mode = s.syncookies_mode
        self.since           = s.since
        self.states          = s.states
        self.states_halfopen = s.states_halfopen
        self.src_nodes       = s.src_nodes
        self.debug           = s.debug
        self.hostid          = ntohl(s.hostid) & 0xffffffff
        self.reass           = s.reass
        self.pf_chksum       = "0x" + "".join(map("{:02x}".format, s.pf_chksum))

        self.cnt       = {'match':                    s.counters[0],
                          'bad-offset':               s.counters[1],
                          'fragment':                 s.counters[2],
                          'short':                    s.counters[3],
                          'normalize':                s.counters[4],
                          'memory':                   s.counters[5],
                          'bad-timestamp':            s.counters[6],
                          'congestion':               s.counters[7],
                          'ip-option':                s.counters[8],
                          'proto-cksum':              s.counters[9],
                          'state-mismatch':           s.counters[10],
                          'state-insert':             s.counters[11],
                          'state-limit':              s.counters[12],
                          'src-limit':                s.counters[13],
                          'synproxy':                 s.counters[14],
                          'translate':                s.counters[15],
                          'no-route':                 s.counters[16]}

        self.lcnt      = {'max states per rule':      s.lcounters[0],
                          'max-src-states':           s.lcounters[1],
                          'max-src-nodes':            s.lcounters[2],
                          'max-src-conn':             s.lcounters[3],
                          'max-src-conn-rate':        s.lcounters[4],
                          'overload table insertion': s.lcounters[5],
                          'overload flush states':    s.lcounters[6],
                          'synfloods detected':       s.lcounters[7],
                          'syncookies sent':          s.lcounters[8],
                          'syncookies validated':     s.lcounters[9]}

        self.fcnt      = {'searches':                 s.fcounters[0],
                          'inserts':                  s.fcounters[1],
                          'removals':                 s.fcounters[2]}

        self.scnt      = {'searches':                 s.scounters[0],
                          'inserts':                  s.scounters[1],
                          'removals':                 s.scounters[2]}

        self.bytes     = {'in':   (s.bcounters[0][0], s.bcounters[1][0]),
                          'out':  (s.bcounters[0][1], s.bcounters[1][1])}

        self.packets   = {'in':  ((s.pcounters[0][0][PF_PASS],
                                   s.pcounters[1][0][PF_PASS]),
                                  (s.pcounters[0][0][PF_DROP],
                                   s.pcounters[1][0][PF_DROP])),
                          'out': ((s.pcounters[0][1][PF_PASS],
                                   s.pcounters[1][1][PF_PASS]),
                                  (s.pcounters[0][1][PF_DROP],
                                   s.pcounters[1][1][PF_DROP]))}

    def _to_string(self):
        """Return a string containing the statistics."""
        s = "Status: " + ('Enabled' if self.running else 'Disabled')

        if self.since:
            runtime = uptime() - self.since
            day, sec = divmod(runtime, 60)
            day, min = divmod(day, 60)
            day, hrs = divmod(day, 24)
            s += " for {} days {:02}:{:02}:{:02}".format(day, hrs, min, sec)

        dbg = next((k for k, v in dbg_levels.items() if v == self.debug),
                   "unknown")
        s = "{:<44}{:>15}\n\n".format(s, "Debug: " + dbg)
        s += "Hostid:   0x{.hostid:08x}\n".format(self)
        s += "Checksum: {.pf_chksum}\n\n".format(self)

        if self.ifname:
            fmt = "  {0:<25} {1[0]:>14d} {1[1]:>16d}\n"
            s += "Interface Stats for {.ifname:<16} ".format(self)
            s += "{:>5} {:>16}\n".format("IPv4", "IPv6")
            s += fmt.format("Bytes In", self.bytes["in"])
            s += fmt.format("Bytes Out", self.bytes["out"])
            s += "  Packets In\n"
            s += fmt.format("  Passed", self.packets["in"][PF_PASS])
            s += fmt.format("  Blocked", self.packets["in"][PF_DROP])
            s += "  Packets Out\n"
            s += fmt.format("  Passed", self.packets["out"][PF_PASS])
            s += fmt.format("  Blocked", self.packets["out"][PF_DROP])
            s += "\n"

        s += "{:<27} {:>14} {:>16}\n".format("State Table", "Total", "Rate")
        s += "  {:<25} {.states:>14d}\n".format("current entries", self)
        s += "  {:<25} {.states_halfopen:>14d}".format("half-open tcp", self)
        for k, v in self.fcnt.items():
            s += "\n  {:<25} {:>14d} ".format(k, v)
            if self.since and runtime:
                s += "{:>14.1f}/s".format(float(v)/runtime)

        s += "\nSource Tracking Table\n"
        s += "  {:<25} {.src_nodes:>14d}".format("current entries", self)
        for k, v in self.scnt.items():
            s += "\n  {:<25} {:>14d} ".format(k, v)
            if self.since and runtime:
                s += "{:>14.1f}/s".format(float(v)/runtime)

        s += "\nCounters"
        for k, v in self.cnt.items():
            s += "\n  {:<25} {:>14d} ".format(k, v)
            if self.since and runtime:
                s += "{:>14.1f}/s".format(float(v)/runtime)

        s += "\nLimit Counters"
        for k, v in self.lcnt.items():
            s += "\n  {:<25} {:>14d} ".format(k, v)
            if self.since and runtime:
                s += "{:>14.1f}/s".format(float(v)/runtime)

        return s


class PFIface(PFObject):
    """Class representing a network interface."""

    _struct_type = pfi_kif

    def __init__(self, iface):
        """Check argument and initialize class attributes."""
        super(PFIface, self).__init__(iface)

    def _from_struct(self, i):
        """Initialize class attributes from a pfi_kif structure."""
        self.name      = i.pfik_name.decode()
        self.packets   = {'in':  ((i.pfik_packets[0][0][PF_PASS],
                                   i.pfik_packets[1][0][PF_PASS]),
                                  (i.pfik_packets[0][0][PF_DROP],
                                   i.pfik_packets[1][0][PF_DROP])),
                          'out': ((i.pfik_packets[0][1][PF_PASS],
                                   i.pfik_packets[1][1][PF_PASS]),
                                  (i.pfik_packets[0][1][PF_DROP],
                                   i.pfik_packets[1][1][PF_DROP]))}
        self.bytes     = {'in':  ((i.pfik_bytes[0][0][PF_PASS],
                                   i.pfik_bytes[1][0][PF_PASS]),
                                  (i.pfik_bytes[0][0][PF_DROP],
                                   i.pfik_bytes[1][0][PF_DROP])),
                          'out': ((i.pfik_bytes[0][1][PF_PASS],
                                   i.pfik_bytes[1][1][PF_PASS]),
                                  (i.pfik_bytes[0][1][PF_DROP],
                                   i.pfik_bytes[1][1][PF_DROP]))}
        self.flags     = i.pfik_flags
        self.flags_new = i.pfik_flags_new
        self.states    = i.pfik_states
        self.cleared   = i.pfik_tzero
        self.rules     = i.pfik_rules
        self.routes    = i.pfik_routes

    def _to_string(self):
        """Return a string containing the description of the interface."""
        if self.flags & PFI_IFLAG_SKIP:
            s = "{.name} (skip)\n".format(self)
        else:
            s = "{.name}\n".format(self)
        s += "\tCleared:     {}\n".format(time.ctime(self.cleared))
        s += "\tReferences:  [ States:  {.states:<18d}".format(self)
        s += " Rules: {.rules:<18d} ]\n".format(self)

        pfik_ops = ("Pass:", "Block:")
        for o, p, b in zip(pfik_ops, self.packets["in"], self.bytes["in"]):
            l = "\tIn4/{:<6s}   [ Packets: {:<18d} Bytes: {:<18d} ]\n"
            s += l.format(o, p[0], b[0])
        for o, p, b in zip(pfik_ops, self.packets["out"], self.bytes["out"]):
            l = "\tOut4/{:<6s}  [ Packets: {:<18d} Bytes: {:<18d} ]\n"
            s += l.format(o, p[0], b[0])
        for o, p, b in zip(pfik_ops, self.packets["in"], self.bytes["in"]):
            l = "\tIn6/{:<6s}   [ Packets: {:<18d} Bytes: {:<18d} ]\n"
            s += l.format(o, p[1], b[1])
        for o, p, b in zip(pfik_ops, self.packets["out"], self.bytes["out"]):
            l = "\tOut6/{:<6s}  [ Packets: {:<18d} Bytes: {:<18d} ]\n"
            s += l.format(o, p[1], b[1])

        return s
